import keyword
import types
from reprlib import recursive_repr

import msgspec
import typing_extensions as typing
from fntypes.co import Nothing

from telegrinder.msgspec_utils import Option, decoder, encoder, struct_asdict

UNSET: typing.Final[typing.Any] = typing.cast("typing.Any", msgspec.UNSET)
"""See [DOCS](https://jcristharif.com/msgspec/api.html#unset) about `msgspec.UNSET`."""
MODEL_CONFIG: typing.Final[dict[str, typing.Any]] = {
    "dict": True,
    "rename": {kw + "_": kw for kw in keyword.kwlist},
}
INSPECTED_MODEL_FIELDS_KEY: typing.Final[str] = "__inspected_struct_fields__"


def is_none(obj: typing.Any, /) -> typing.TypeIs[Nothing | None]:
    return isinstance(obj, Nothing | types.NoneType)


if typing.TYPE_CHECKING:

    @typing.overload
    def field() -> typing.Any: ...

    @typing.overload
    def field(*, name: str | None = ...) -> typing.Any: ...

    @typing.overload
    def field(*, default: typing.Any, name: str | None = ...) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        default_factory: typing.Callable[[], typing.Any],
        name: str | None = None,
    ) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        converter: typing.Callable[[typing.Any], typing.Any],
        name: str | None = ...,
    ) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        default: typing.Any,
        converter: typing.Callable[[typing.Any], typing.Any],
        name: str | None = ...,
    ) -> typing.Any: ...

    @typing.overload
    def field(
        *,
        default_factory: typing.Callable[[], typing.Any],
        converter: typing.Callable[[typing.Any], typing.Any],
        name: str | None = None,
    ) -> typing.Any: ...

    def field(
        *,
        default=...,
        default_factory=...,
        name=...,
        converter=...,
    ) -> typing.Any: ...

    class From[T]:
        def __new__(cls, _: T, /) -> typing.Any: ...
else:
    from msgspec import field as _field

    type From[T] = T

    def field(**kwargs):
        if (default := kwargs.get("default")) is Ellipsis:
            kwargs["default"] = UNSET

        kwargs.pop("converter", None)
        return _field(**kwargs)


@typing.dataclass_transform(field_specifiers=(field,))
class Model(msgspec.Struct, **MODEL_CONFIG):
    if not typing.TYPE_CHECKING:

        def __getattribute__(self, name, /):
            class_ = type(self)
            val = object.__getattribute__(self, name)

            if name not in class_.__struct_fields__:
                return val

            if (
                (field_info := class_.get_fields().get(name)) is not None
                and isinstance(field_info.type, msgspec.inspect.CustomType)
                and issubclass(field_info.type.cls, Option)
            ):
                return Nothing() if val is UNSET else val

            if val is UNSET:
                raise AttributeError(f"{class_.__name__!r} object has no attribute {name!r}")

            return val

    def __post_init__(self) -> None:
        for field, value in struct_asdict(self, exclude_unset=False).items():
            if is_none(value):
                setattr(self, field, UNSET)

    @recursive_repr()
    def __repr__(self) -> str:
        return "{}({})".format(
            type(self).__name__,
            ", ".join(
                f"{f}={val!r}"
                for f, val in struct_asdict(self, exclude_unset=False, unset_as_nothing=True).items()
            ),
        )

    @classmethod
    def get_fields(cls) -> types.MappingProxyType[str, msgspec.inspect.Field]:
        if (model_fields := getattr(cls, INSPECTED_MODEL_FIELDS_KEY, None)) is not None:
            return model_fields

        model_fields = types.MappingProxyType[str, msgspec.inspect.Field](
            {f.name: f for f in msgspec.inspect.type_info(cls).fields},  # type: ignore
        )
        setattr(cls, INSPECTED_MODEL_FIELDS_KEY, model_fields)
        return model_fields

    @classmethod
    def from_data[**P, T](cls: typing.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        return decoder.convert(msgspec.structs.asdict(cls(*args, **kwargs)), type=cls)  # type: ignore

    @classmethod
    def from_dict(cls, obj: dict[str, typing.Any], /) -> typing.Self:
        return decoder.convert(obj, type=cls)

    @classmethod
    def from_raw(cls, raw: str | bytes, /) -> typing.Self:
        return decoder.decode(raw, type=cls)

    def _to_dict(
        self,
        dct_name: str,
        exclude_fields: set[str],
        full: bool,
    ) -> dict[str, typing.Any]:
        if dct_name not in self.__dict__:
            self.__dict__[dct_name] = (
                struct_asdict(self)
                if not full
                else encoder.to_builtins(self.to_dict(exclude_fields=exclude_fields), order="deterministic")
            )

        if not exclude_fields:
            return self.__dict__[dct_name]

        return {key: value for key, value in self.__dict__[dct_name].items() if key not in exclude_fields}

    def to_raw(self) -> str:
        return encoder.encode(self)

    def to_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]:
        return self._to_dict("model_as_dict", exclude_fields or set(), full=False)

    def to_full_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]:
        return self._to_dict("model_as_full_dict", exclude_fields or set(), full=True)


__all__ = ("MODEL_CONFIG", "UNSET", "Model", "field", "is_none")



from telegrinder import API, Choice, Message, Telegrinder, Token, WaiterMachine
from telegrinder.bot.dispatch.waiter_machine.hasher.callback import CALLBACK_QUERY_FOR_MESSAGE
from telegrinder.modules import logger
from telegrinder.rules import Text

api = API(token=Token.from_env())
bot = Telegrinder(api)
wm = WaiterMachine(bot.dispatch)

logger.set_level("INFO")


@bot.on.message(Text("/choice"))
async def action(m: Message):
    chosen, m_id = await (
        Choice(wm, m.chat.id, "Choose something", max_in_row=1)
        .add_option("apple", "Apple 🔴", "Apple 🟢")
        .add_option("banana", "Banana 🔴", "Banana 🟢", is_picked=True)
        .add_option("pear", "Pear 🔴", "Pear 🟢")
        .wait(CALLBACK_QUERY_FOR_MESSAGE, m.ctx_api)
    )
    await m.edit(text=f"You chose - {chosen}", message_id=m_id)


bot.run_forever()



import msgspec

from telegrinder import (
    API,
    CallbackQuery,
    InlineButton,
    InlineKeyboard,
    Message,
    StaticButton,
    StaticKeyboard,
    Telegrinder,
    Token,
)
from telegrinder.modules import logger
from telegrinder.rules import PayloadModelRule, Text

logger.set_level("DEBUG")
api = API(token=Token.from_env())
bot = Telegrinder(api)


class FruitsKeyboard(StaticKeyboard, max_in_row=2, one_time_keyboard=True):
    APPLE = StaticButton("🍏 Apple")
    BANANA = StaticButton("Banana 🍌")
    GRAPES = StaticButton("🍇 Grapes")
    KIWI = StaticButton("Kiwi 🥝")


# Alternative to msgspec.Struct: use @dataclasses.dataclass decorator
class ItemModel(msgspec.Struct):
    item: str
    amount: int
    action: str


kb = (
    InlineKeyboard()
    .add(
        InlineButton(
            text="buy a doughnut",
            callback_data=ItemModel(item="doughnut", amount=100, action="buy"),
        )
    )
    .add(InlineButton(text="buy a cake", callback_data=ItemModel(item="cake", amount=1000, action="buy")))
).get_markup()


@bot.on.message(Text("/start"))
async def start(message: Message) -> None:
    await message.answer(
        text="Hello! Choose what you need:",
        reply_markup=kb,
    )


@bot.on.message(Text("/eat"))
async def eat_some_fruit(message: Message) -> None:
    await message.reply(
        text="What fruit do you want to eat?",
        reply_markup=FruitsKeyboard.get_markup(),
    )


@bot.on.message(FruitsKeyboard.APPLE)
async def eat_apple(message: Message) -> None:
    await message.reply(
        text="Very nice green apple, bon appetit!",
        reply_markup=FruitsKeyboard.get_keyboard_remove(),
    )


@bot.on.message(FruitsKeyboard.BANANA)
async def eat_banana(message: Message) -> None:
    await message.reply(
        text="This banana is so ripe and sweet, bon appetit!",
        reply_markup=FruitsKeyboard.get_keyboard_remove(),
    )


@bot.on.message(FruitsKeyboard.GRAPES)
async def eat_grapes(message: Message) -> None:
    await message.reply(
        text="Grapes with seeds, bon appetit!",
        reply_markup=FruitsKeyboard.get_keyboard_remove(),
    )


@bot.on.message(FruitsKeyboard.KIWI)
async def eat_kiwi(message: Message) -> None:
    await message.reply(
        text="Very sour kiwi... Bon appetit!",
        reply_markup=FruitsKeyboard.get_keyboard_remove(),
    )


@bot.on.callback_query(PayloadModelRule(ItemModel, alias="data"))
async def buy(cb: CallbackQuery, data: ItemModel) -> None:
    await cb.edit_text(f"You bought a {data.item} for {data.amount}")


bot.run_forever(skip_updates=True)



from __future__ import annotations

import dataclasses
import typing
from collections import deque
from functools import cached_property

from fntypes.result import Error, Ok, Result

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.abc import ABCHandler
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.modules import logger
from telegrinder.node.base import IsNode, get_nodes
from telegrinder.node.composer import compose_nodes
from telegrinder.tools.fullname import fullname
from telegrinder.tools.magic.function import bundle
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from telegrinder.bot.rules.abc import ABCRule

type Function = typing.Callable[..., typing.Coroutine[typing.Any, typing.Any, typing.Any]]


@dataclasses.dataclass(repr=False, slots=True)
class FuncHandler[T: Function](ABCHandler):
    function: T
    rules: dataclasses.InitVar[typing.Iterable[ABCRule] | None] = None
    final: bool = dataclasses.field(default=True, kw_only=True)
    preset_context: Context = dataclasses.field(default_factory=lambda: Context(), kw_only=True)

    def __post_init__(self, rules: typing.Iterable[ABCRule] | None) -> None:
        self.check_rules = deque(rules or ())

    @property
    def __call__(self) -> Function:
        return self.function

    def __repr__(self) -> str:
        return fullname(self.function)

    @cached_property
    def required_nodes(self) -> dict[str, IsNode]:
        return get_nodes(self.function)

    async def run(
        self,
        api: API,
        event: Update,
        context: Context,
        check: bool = True,
    ) -> Result[typing.Any, str]:
        logger.debug("Checking rules and composing nodes for handler `{!r}`...", self)

        temp_ctx = context.copy()
        temp_ctx |= self.preset_context.copy()

        if check:
            for rule in self.check_rules:
                if not await check_rule(api, rule, event, temp_ctx):
                    return Error(f"Rule {rule!r} failed.")

        context |= temp_ctx
        data = {Update: event, API: api}
        node_col = None

        if self.required_nodes:
            match await compose_nodes(self.required_nodes, context, data=data):
                case Ok(value):
                    node_col = value
                case Error(compose_error):
                    return Error(f"Cannot compose nodes for handler `{self}`, error: {compose_error.message}")

        logger.debug("All good, running handler `{!r}`", self)

        temp_ctx = context.copy()
        try:
            bundle_function = bundle(
                self.function,
                {**data, Context: temp_ctx},
                typebundle=True,
                start_idx=0,
            )
            bundle_function &= bundle(
                self.function, context | ({} if node_col is None else node_col.values), start_idx=0
            )
            return Ok(await bundle_function())
        finally:
            context |= temp_ctx

            if node_col is not None:
                await node_col.close_all()


__all__ = ("FuncHandler",)



import typing
from contextlib import contextmanager

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.bot.rules.abc import ABCRule, check_rule
from telegrinder.node.base import IsNode
from telegrinder.node.composer import compose_nodes
from telegrinder.types import Update
from telegrinder.types.objects import Update


class GlobalMiddleware(ABCMiddleware):
    def __init__(self) -> None:
        self.filters: set[ABCRule] = set()
        self.source_filters: dict[IsNode, dict[typing.Any, ABCRule]] = {}

    async def pre(self, event: Update, api: API, ctx: Context) -> bool:
        for filter in self.filters:
            if not await check_rule(api, filter, event, ctx):
                return False

        # Simple implication. Grouped by source categories
        for source, identifiers in self.source_filters.items():
            result = await compose_nodes({"value": source}, ctx, {Update: event, API: api})
            if result := result.unwrap():
                result = result.values["value"]
            else:
                return True

            if result in identifiers:
                return await check_rule(api, identifiers[result], event, ctx)

        return True

    @contextmanager
    def apply_filters(
        self,
        *filters: ABCRule,
        source_filter: tuple[IsNode, typing.Any, ABCRule] | None = None,
    ) -> typing.Generator[None, typing.Any, None]:
        if source_filter is not None:
            self.source_filters.setdefault(source_filter[0], {})
            self.source_filters[source_filter[0]].update({source_filter[1]: source_filter[2]})

        self.filters |= set(filters)
        yield
        self.filters.difference_update(filters)

        if source_filter is not None and (identifiers := self.source_filters.get(source_filter[0])):
            identifiers.pop(source_filter[1], None)


__all__ = ("GlobalMiddleware",)



from telegrinder import API, InlineQuery, Telegrinder, Token
from telegrinder.rules import InlineQueryText
from telegrinder.types import (
    InlineQueryResultArticle,
    InputTextMessageContent,
)

api = API(token=Token.from_env())
bot = Telegrinder(api)


@bot.on.inline_query(InlineQueryText("test"))
async def test_inline(q: InlineQuery):
    await q.answer(
        InlineQueryResultArticle(
            "Press me",
            InputTextMessageContent(message_text="I tested inline query"),
        ),
    )


@bot.on.inline_query()
async def reverse_inline(q: InlineQuery):
    if not q.query:
        return
    await q.answer(
        InlineQueryResultArticle(
            "Send reversed",
            InputTextMessageContent(message_text=q.query[::-1]),
        ),
    )


bot.run_forever()



import dataclasses
import typing

from fntypes.option import Nothing, Option, Some

import telegrinder.types
from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.node.base import ComposeError, DataNode, scalar_node

type AttachmentType = typing.Literal[
    "audio",
    "animation",
    "document",
    "photo",
    "poll",
    "voice",
    "video",
    "video_note",
    "successful_payment",
]


@dataclasses.dataclass(slots=True)
class Attachment(DataNode):
    attachment_type: AttachmentType

    animation: Option[telegrinder.types.Animation] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    audio: Option[telegrinder.types.Audio] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    document: Option[telegrinder.types.Document] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    photo: Option[list[telegrinder.types.PhotoSize]] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    poll: Option[telegrinder.types.Poll] = dataclasses.field(default_factory=lambda: Nothing(), kw_only=True)
    voice: Option[telegrinder.types.Voice] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    video: Option[telegrinder.types.Video] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    video_note: Option[telegrinder.types.VideoNote] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )
    successful_payment: Option[telegrinder.types.SuccessfulPayment] = dataclasses.field(
        default_factory=Nothing,
        kw_only=True,
    )

    @classmethod
    def get_attachment_types(cls) -> tuple[AttachmentType, ...]:
        return typing.get_args(AttachmentType.__value__)

    @classmethod
    def compose(cls, message: MessageCute) -> typing.Self:
        for attachment_type in cls.get_attachment_types():
            match getattr(message, attachment_type, Nothing()):
                case Some(attachment):
                    return cls(attachment_type, **{attachment_type: Some(attachment)})

        raise ComposeError("No attachment found in message.")


@dataclasses.dataclass(slots=True)
class Photo(DataNode):
    sizes: list[telegrinder.types.PhotoSize]

    @classmethod
    def compose(cls, attachment: Attachment) -> typing.Self:
        if not attachment.photo:
            raise ComposeError("Attachment is not a photo.")
        return cls(attachment.photo.unwrap())


@scalar_node
class Video:
    @classmethod
    def compose(cls, attachment: Attachment) -> telegrinder.types.Video:
        if not attachment.video:
            raise ComposeError("Attachment is not a video.")
        return attachment.video.unwrap()


@scalar_node
class VideoNote:
    @classmethod
    def compose(cls, attachment: Attachment) -> telegrinder.types.VideoNote:
        if not attachment.video_note:
            raise ComposeError("Attachment is not a video note.")
        return attachment.video_note.unwrap()


@scalar_node
class Audio:
    @classmethod
    def compose(cls, attachment: Attachment) -> telegrinder.types.Audio:
        if not attachment.audio:
            raise ComposeError("Attachment is not an audio.")
        return attachment.audio.unwrap()


@scalar_node
class Animation:
    @classmethod
    def compose(cls, attachment: Attachment) -> telegrinder.types.Animation:
        if not attachment.animation:
            raise ComposeError("Attachment is not an animation.")
        return attachment.animation.unwrap()


@scalar_node
class Voice:
    @classmethod
    def compose(cls, attachment: Attachment) -> telegrinder.types.Voice:
        if not attachment.voice:
            raise ComposeError("Attachment is not a voice.")
        return attachment.voice.unwrap()


@scalar_node
class Document:
    @classmethod
    def compose(cls, attachment: Attachment) -> telegrinder.types.Document:
        if not attachment.document:
            raise ComposeError("Attachment is not a document.")
        return attachment.document.unwrap()


@scalar_node
class Poll:
    @classmethod
    def compose(cls, attachment: Attachment) -> telegrinder.types.Poll:
        if not attachment.poll:
            raise ComposeError("Attachment is not a poll.")
        return attachment.poll.unwrap()


@scalar_node
class SuccessfulPayment:
    @classmethod
    def compose(cls, attachment: Attachment) -> telegrinder.types.SuccessfulPayment:
        if not attachment.successful_payment:
            raise ComposeError("Attachment is not a successful payment.")
        return attachment.successful_payment.unwrap()


__all__ = (
    "Animation",
    "Attachment",
    "Audio",
    "Document",
    "Photo",
    "Poll",
    "SuccessfulPayment",
    "Video",
    "VideoNote",
    "Voice",
)



from telegrinder import API, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.node.i18n import BaseTranslator, I18NConfig, KeySeparator
from telegrinder.node.source import UserSource
from telegrinder.rules import Text

bot = Telegrinder(API(Token.from_env()))

logger.set_level("INFO")
BaseTranslator.configure(I18NConfig(domain="messages", folder="examples/assets/i18n"))
KeySeparator.set(" ")


@bot.on.message(Text("hi"))
async def hi(_: BaseTranslator) -> str:
    return _.hi()


@bot.on.message(Text("hello"))
async def hello(_: BaseTranslator, user: UserSource) -> str:
    return _("Hello, {name}!", name=user.full_name)


@bot.on.message(Text("how are you?"))
async def how_are_you(_: BaseTranslator) -> str:
    return _.im.fine()


bot.run_forever()



from telegrinder.bot.dispatch.return_manager.pre_checkout_query import PreCheckoutQueryManager
from telegrinder.bot.dispatch.view.base import BaseView
from telegrinder.types.enums import UpdateType


class PreCheckoutQueryView(BaseView):
    def __init__(self) -> None:
        super().__init__(UpdateType.PRE_CHECKOUT_QUERY)
        self.return_manager = PreCheckoutQueryManager()


__all__ = ("PreCheckoutQueryView",)



import typing

from telegrinder import API, Telegrinder, Token
from telegrinder.node.event import EventNode
from telegrinder.rules import Text
from telegrinder.types.objects import Message

bot = Telegrinder(API(Token.from_env()))


# bot.on() is the same as bot.on.raw_event()
@bot.on(Text(["hello", "hi"], ignore_case=True))
async def handle_raw_message(raw_msg: EventNode[Message]) -> None:
    await bot.api.send_message(chat_id=raw_msg.chat_id, text="Hello, World!")


@bot.on()
async def handle_raw_cb(event: EventNode[dict[str, typing.Any]]) -> None:
    await bot.api.answer_callback_query(callback_query_id=event["id"])


bot.run_forever()



import datetime as dt
import typing

from telegrinder.msgspec_utils.abc import SupportsCast


class datetime(dt.datetime, SupportsCast):  # noqa: N801  # type: ignore
    @classmethod
    def cast(cls, obj: dt.datetime) -> typing.Self:
        return cls.fromtimestamp(timestamp=obj.timestamp(), tz=obj.tzinfo)


class timedelta(dt.timedelta, SupportsCast):  # noqa: N801  # type: ignore
    @classmethod
    def cast(cls, obj: dt.timedelta) -> typing.Self:
        return cls(seconds=obj.total_seconds())


if typing.TYPE_CHECKING:
    datetime: typing.TypeAlias = dt.datetime
    timedelta: typing.TypeAlias = dt.timedelta


__all__ = ("datetime", "timedelta")



from __future__ import annotations

import dataclasses
import enum
import inspect
import types
import typing
from functools import cached_property, wraps

from telegrinder.tools.magic.annotations import Annotations

type Function[**P, R] = typing.Callable[P, R]
type AnyFunction = Function[..., typing.Any]


def _to_str(obj: typing.Any, /) -> str:
    if isinstance(obj, enum.Enum):
        return str(obj.value)
    return str(obj) if not isinstance(obj, str) else obj


def _resolve_arg_names(
    func: AnyFunction,
    /,
    *,
    start_idx: int,
    stop_idx: int,
    exclude: set[str] | None = None,
) -> tuple[str, ...]:
    exclude = exclude or set()
    varnames = func.__code__.co_varnames[start_idx:stop_idx]
    return tuple(name for name in varnames if name not in exclude)


class FunctionParameters(typing.TypedDict):
    args: list[tuple[str, typing.Any | inspect.Parameter.empty]]
    kwargs: list[tuple[str, typing.Any | inspect.Parameter.empty]]
    var_star_args: typing.NotRequired[str]
    var_star_kwargs: typing.NotRequired[str]


@dataclasses.dataclass(frozen=True, repr=False)
class Bundle[R]:
    function: Function[..., R]
    start_idx: int
    context: types.MappingProxyType[str, typing.Any]

    @cached_property
    def args(self) -> tuple[typing.Any, ...]:
        return tuple(
            self.context[name]
            for name in resolve_posonly_arg_names(self.function, start_idx=self.start_idx)
            if name in self.context
        )

    @cached_property
    def kwargs(self) -> dict[str, typing.Any]:
        posonly_arg_names = resolve_posonly_arg_names(self.function, start_idx=self.start_idx)
        return {name: value for name, value in self.context.items() if name not in posonly_arg_names}

    @classmethod
    def from_context(
        cls,
        function: Function[..., R],
        context: typing.Mapping[str, typing.Any],
        /,
        *,
        start_idx: int = 1,
    ) -> typing.Self:
        return cls(function, start_idx, types.MappingProxyType(context))

    def __repr__(self) -> str:
        return "<{}: function={!r} bundle args={!r} kwargs={!r}>".format(
            type(self).__name__,
            self.function,
            self.args,
            self.kwargs,
        )

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> R:
        return self.function(*args, *self.args, **kwargs, **self.kwargs)

    def __and__(self, other: object, /) -> typing.Self:
        if not isinstance(other, Bundle):
            return NotImplemented

        if self.function != other.function:
            raise ValueError(
                f"Cannot merge contexts because bundle {other!r} is intended for a different function.",
            )

        return dataclasses.replace(
            self,
            context=types.MappingProxyType(mapping=self.context | other.context),
        )

    def __iand__(self, other: object, /) -> typing.Self:
        return self.__and__(other)


def function_context[**P, R](key: str, /) -> Function[[Function[P, R]], Function[P, R]]:
    @lambda wrapper: typing.cast("Function[[Function[P, R]], Function[P, R]]", wrapper)
    def wrapper(func: Function[typing.Concatenate[AnyFunction, P], R], /) -> AnyFunction:
        @wraps(func)
        def inner(passed_function: AnyFunction, /, *args: P.args, **kwargs: P.kwargs) -> R:
            sentinel = object()
            context: dict[str, typing.Any] = passed_function.__dict__.setdefault("__function_context__", {})

            if (value := context.get(key, sentinel)) is not sentinel:
                return value

            context[key] = result = func(passed_function, *args, **kwargs)
            return result

        return inner

    return wrapper


def resolve_arg_names(
    func: AnyFunction,
    /,
    *,
    start_idx: int = 1,
    exclude: set[str] | None = None,
) -> tuple[str, ...]:
    return _resolve_arg_names(
        func,
        start_idx=start_idx,
        stop_idx=func.__code__.co_argcount + func.__code__.co_kwonlyargcount,
        exclude=exclude,
    )


def resolve_kwonly_arg_names(
    func: AnyFunction,
    /,
    *,
    start_idx: int = 1,
    exclude: set[str] | None = None,
) -> tuple[str, ...]:
    return _resolve_arg_names(
        func,
        start_idx=func.__code__.co_argcount + start_idx,
        stop_idx=func.__code__.co_argcount + func.__code__.co_kwonlyargcount,
        exclude=exclude,
    )


def resolve_posonly_arg_names(
    func: AnyFunction,
    /,
    *,
    start_idx: int = 1,
    exclude: set[str] | None = None,
) -> tuple[str, ...]:
    return _resolve_arg_names(
        func,
        start_idx=start_idx,
        stop_idx=func.__code__.co_posonlyargcount,
        exclude=exclude,
    )


@function_context("default_arguments")
def get_default_args(func: AnyFunction, /) -> dict[str, typing.Any]:
    parameters = get_func_parameters(func)
    return {
        key: value
        for sublist in (parameters["args"], parameters["kwargs"])
        for key, value in sublist
        if value is not inspect.Parameter.empty
    }


@function_context("signature_parameters")
def get_func_parameters(func: AnyFunction, /) -> FunctionParameters:
    func_params = FunctionParameters(args=[], kwargs=[])

    for k, p in inspect.signature(func).parameters.items():
        if k in ("self", "cls"):
            continue

        match p.kind:
            case p.POSITIONAL_OR_KEYWORD | p.POSITIONAL_ONLY:
                func_params["args"].append((k, p.default))
            case p.KEYWORD_ONLY:
                func_params["kwargs"].append((k, p.default))
            case p.VAR_POSITIONAL:
                func_params["var_star_args"] = k
            case p.VAR_KEYWORD:
                func_params["var_star_kwargs"] = k

    return func_params


@function_context("resolved_annotations")
def get_func_annotations(func: AnyFunction, /) -> typing.Mapping[str, typing.Any]:
    return Annotations(obj=func).get(
        ignore_failed_evals=True,
        exclude_forward_refs=True,
        allow_return_type=False,
        cache=False,
    )


def bundle[R](
    function: Function[..., R],
    context: dict[typing.Any, typing.Any],
    /,
    *,
    start_idx: int = 1,
    bundle_kwargs: bool = False,
    typebundle: bool = False,
    omit_defaults: bool = False,
) -> Bundle[R]:
    if typebundle:
        return Bundle.from_context(
            function,
            {name: context[ann] for name, ann in get_func_annotations(function).items() if ann in context},
            start_idx=start_idx,
        )

    kwargs = {_to_str(k): v for k, v in context.items()}
    if not bundle_kwargs or "var_star_kwargs" not in get_func_parameters(function):
        names = resolve_arg_names(function, start_idx=start_idx, exclude={"cls", "self"})
        kwargs = {k: v for k, v in kwargs.items() if k in names}

    return Bundle.from_context(
        function,
        kwargs if omit_defaults else get_default_args(function) | kwargs,
        start_idx=start_idx,
    )


__all__ = (
    "Bundle",
    "bundle",
    "function_context",
    "get_default_args",
    "get_func_annotations",
    "get_func_parameters",
    "resolve_arg_names",
    "resolve_kwonly_arg_names",
    "resolve_posonly_arg_names",
)



from __future__ import annotations

import abc
import typing


def is_supports_cast(obj: typing.Any, /) -> typing.TypeGuard[SupportsCast]:
    return isinstance(obj, SupportsCast)


@typing.runtime_checkable
class SupportsCast(typing.Protocol):
    @classmethod
    @abc.abstractmethod
    def cast(cls, obj: typing.Any) -> typing.Self: ...


__all__ = ("SupportsCast", "is_supports_cast")



from telegrinder.msgspec_utils.abc import SupportsCast, is_supports_cast
from telegrinder.msgspec_utils.custom_types.datetime import datetime, timedelta
from telegrinder.msgspec_utils.custom_types.enum_meta import BaseEnumMeta
from telegrinder.msgspec_utils.custom_types.literal import Literal
from telegrinder.msgspec_utils.custom_types.option import Option
from telegrinder.msgspec_utils.decoder import Decoder, convert, decoder
from telegrinder.msgspec_utils.encoder import Encoder, encoder, to_builtins
from telegrinder.msgspec_utils.json import dumps, loads
from telegrinder.msgspec_utils.tools import (
    get_class_annotations,
    get_origin,
    get_type_hints,
    is_common_type,
    struct_asdict,
    type_check,
)

__all__ = (
    "BaseEnumMeta",
    "Decoder",
    "Encoder",
    "Literal",
    "Option",
    "SupportsCast",
    "convert",
    "datetime",
    "decoder",
    "dumps",
    "encoder",
    "get_class_annotations",
    "get_origin",
    "get_type_hints",
    "is_common_type",
    "is_supports_cast",
    "loads",
    "struct_asdict",
    "timedelta",
    "to_builtins",
    "type_check",
)



import typing

from telegrinder import API, Message, Telegrinder, Token
from telegrinder.bot.dispatch.action import action
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.bot.rules.text import Text
from telegrinder.node import ChatId, Error, UserId

bot = Telegrinder(API(Token.from_env()))


def is_not_admin() -> typing.NoReturn:
    # some pretty cool logic here...
    raise IsNotAdminError("you are not admin... not like a boss... im so sori 😞")


class IsNotAdminError(Exception):
    pass


class IsAdmin(ABCRule):
    async def check(self, chat_id: ChatId, user_id: UserId) -> bool:
        for chat_member in (await bot.api.get_chat_administrators(chat_id=chat_id)).unwrap():
            if chat_member.v.user.id == user_id and chat_member.v.status in {"creator", "administrator"}:
                return True

        return False


@bot.on.message(Text("im boss"))
@action(is_not_admin, when=~IsAdmin())
async def handler(message: Message) -> None:
    await message.answer("hi boss 😎")


@bot.on.error()
async def handle_not_admin_error(message: Message, e: Error[IsNotAdminError]) -> None:
    await message.answer(str(e.exception))


bot.run_forever(skip_updates=True)



from msgspec import Struct, field


class Model(Struct):
    pass


class TelegramBotAPISchema(Model, rename={"objects": "types"}):
    version: str
    """API version."""

    release_date: str
    """Release date."""

    changelog: str
    """Changelog."""

    methods: list["MethodSchema"] = field(default_factory=lambda: [])
    """Schema of methods."""

    objects: list["ObjectSchema"] = field(default_factory=lambda: [])
    """Schema of objects."""


class MethodSchema(Model, rename={"params": "fields"}):
    name: str
    """Method name."""

    href: str
    """Url of the documentation about method."""

    returns: list[str] = field(default_factory=lambda: [])
    """Return type."""

    description: list[str] = field(default_factory=lambda: [])
    """Method description."""

    params: list["MethodParameter"] = field(default_factory=lambda: [])
    """Method parameters."""


class ObjectSchema(Model):
    name: str
    """Object name."""

    href: str
    """Url of the documentation about object."""

    description: list[str] = field(default_factory=lambda: [])
    """Object description."""

    fields: list["ObjectField"] = field(default_factory=lambda: [])
    """Object fields."""

    subtypes: list[str] = field(default_factory=lambda: [])
    """Object is inherited by other objects."""

    subtype_of: list[str] = field(default_factory=lambda: [])
    """List of objects that object inherits."""


class ObjectField(Model):
    name: str
    """Field name."""

    types: list[str]
    """Field types."""

    required: bool = False
    """True if field is required."""

    description: str | None = None
    """Field description."""

    default: str | None = None
    """Default value."""

    default_factory: str | None = None
    """Default factory for dataclass."""


class MethodParameter(Model):
    name: str
    """Parameter name."""

    types: list[str]
    """Parameter types."""

    required: bool = False
    """True if parameter is required."""

    description: str | None = None
    """Parameter description."""


__all__ = (
    "MethodParameter",
    "MethodSchema",
    "Model",
    "ObjectField",
    "ObjectSchema",
    "TelegramBotAPISchema",
)



import typing

from fntypes.result import Result

from telegrinder.api.error import APIError
from telegrinder.bot.cute_types.base import BaseCute, compose_method_params, shortcut
from telegrinder.types.methods_utils import get_params
from telegrinder.types.objects import PreCheckoutQuery, User


class PreCheckoutQueryCute(BaseCute[PreCheckoutQuery], PreCheckoutQuery, kw_only=True):
    @property
    def from_user(self) -> User:
        return self.from_

    @shortcut("answer_pre_checkout_query", custom_params={"pre_checkout_query_id"})
    async def answer(
        self,
        ok: bool,
        *,
        error_message: str | None = None,
        pre_checkout_query_id: str | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.answer_pre_checkout_query()`, see the [documentation](https://core.telegram.org/bots/api#answerprecheckoutquery)

        Once the user has confirmed their payment and shipping details, the Bot
        API sends the final confirmation in the form of an Update with the field pre_checkout_query.
        Use this method to respond to such pre-checkout queries. On success, True
        is returned. Note: The Bot API must receive an answer within 10 seconds after
        the pre-checkout query was sent."""
        params = compose_method_params(
            get_params(locals()), self, default_params={("pre_checkout_query_id", "id")}
        )
        return await self.ctx_api.answer_pre_checkout_query(**params)


__all__ = ("PreCheckoutQueryCute",)



import typing

from telegrinder.tools.final import Final


@typing.final
class ParseMode(Final):
    MARKDOWNV2: typing.Final[str] = "MarkdownV2"
    HTML: typing.Final[str] = "HTML"


__all__ = ("ParseMode",)



import asyncio
import dataclasses
import datetime
import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.tools.aio import cancel_future

if typing.TYPE_CHECKING:
    from telegrinder.bot.cute_types.base import BaseCute

    from .actions import WaiterActions


class ShortStateContext[Event: BaseCute[typing.Any]](typing.NamedTuple):
    event: Event
    context: Context


@dataclasses.dataclass(slots=True)
class ShortState[Event: BaseCute[typing.Any]]:
    event: asyncio.Event
    actions: "WaiterActions[Event]"

    release: ABCRule | None = dataclasses.field(
        default=None,
        kw_only=True,
    )
    filter: ABCRule | None = dataclasses.field(
        default=None,
        kw_only=True,
    )

    lifetime: dataclasses.InitVar[datetime.timedelta | None] = dataclasses.field(
        default=None,
        kw_only=True,
    )

    expiration_date: datetime.datetime | None = dataclasses.field(init=False, kw_only=True)
    creation_date: datetime.datetime = dataclasses.field(init=False)
    context: ShortStateContext[Event] | None = dataclasses.field(default=None, init=False, kw_only=True)

    def __post_init__(self, expiration: datetime.timedelta | None = None) -> None:
        self.creation_date = datetime.datetime.now()
        self.expiration_date = (self.creation_date + expiration) if expiration is not None else None

    async def cancel(self) -> None:
        """Cancel schedule waiters."""
        waiters = typing.cast(
            "typing.Iterable[asyncio.Future[typing.Any]]",
            self.event._waiters,  # type: ignore
        )
        for future in waiters:
            await cancel_future(future)


__all__ = ("ShortState", "ShortStateContext")



from __future__ import annotations

import asyncio
import sys
import typing
from http import HTTPStatus

import msgspec
from fntypes.misc import is_ok

from telegrinder.api.api import API
from telegrinder.api.error import APIServerError, InvalidTokenError
from telegrinder.bot.polling.abc import ABCPolling
from telegrinder.modules import logger
from telegrinder.msgspec_utils import decoder
from telegrinder.tools.aio import maybe_awaitable
from telegrinder.types.objects import Update, UpdateType

DEFAULT_OFFSET: typing.Final[int] = 0
DEFAULT_RECONNECT_AFTER: typing.Final[float] = 5.0
DEFAULT_MAX_RECONNECTS: typing.Final[int] = 15


def _compute_number(
    default: int | float,
    input_value: int | float,
    conditional_value: int | float,
    /,
) -> int | float:
    return max(default, input_value) * (input_value <= conditional_value) + input_value * (
        input_value >= conditional_value
    )


class PollingErrorHandler:
    _handlers: dict[type[BaseException], typing.Callable[[BaseException], typing.Any]]

    __slots__ = ("_polling", "_handlers")

    def __init__(self, polling: Polling, /) -> None:
        self._polling = polling
        self._handlers = {  # type: ignore
            InvalidTokenError: self._handle_invalid_token_error,
            asyncio.CancelledError: self._handle_cancelled_error,
            APIServerError: self._handle_api_server_error,
            **{e: self._handle_connection_timeout_error for e in polling.api.http.CONNECTION_TIMEOUT_ERRORS},
            **{e: self._handle_client_connection_error for e in polling.api.http.CLIENT_CONNECTION_ERRORS},
        }

    async def handle(self, error: BaseException) -> bool:
        error_class = type(error)

        if error_class is SystemExit:
            self._handle_system_exit(error)  # type: ignore

        if error_class not in self._handlers:
            return False

        try:
            await maybe_awaitable(self._handlers[error_class](error))
            return True
        except SystemExit as sys_exit_err:
            self._handle_system_exit(sys_exit_err)

    def _handle_system_exit(self, error: SystemExit) -> typing.NoReturn:
        logger.error(f"Forced exit from the program with code {error.code}.")
        raise error from None

    def _handle_invalid_token_error(
        self,
        error: InvalidTokenError,
    ) -> typing.NoReturn:
        logger.error(error)
        self._polling.stop()
        sys.exit(3)

    def _handle_cancelled_error(self, _: asyncio.CancelledError) -> None:
        logger.info("Caught cancel, stopping polling...")
        self._polling.stop()

    async def _handle_connection_timeout_error(self, _: BaseException) -> None:
        if self._polling.reconnects_counter > self._polling.max_reconnects:
            logger.error(
                "Failed to reconnect to Telegram API server after {} attempts, stopping polling...",
                self._polling.max_reconnects,
            )
            self._polling.stop()
            sys.exit(6)

        logger.warning(
            "Server disconnected, waiting {} seconds to reconnect...",
            self._polling.reconnect_after,
        )
        await asyncio.sleep(self._polling.reconnect_after)

    async def _handle_client_connection_error(self, _: BaseException) -> None:
        logger.error(
            "Client connection failed, attempt to reconnect after {} seconds...",
            self._polling.reconnect_after,
        )
        await asyncio.sleep(self._polling.reconnect_after)

    async def _handle_api_server_error(
        self,
        error: APIServerError,
    ) -> None:
        logger.error(f"{error}, waiting {error.retry_after} seconds to the next request...")
        await asyncio.sleep(error.retry_after)


class Polling(ABCPolling):
    __slots__ = (
        "api",
        "timeout",
        "limit",
        "allowed_updates",
        "reconnect_after",
        "max_reconnects",
        "offset",
        "_running",
        "_reconnects_counter",
        "_error_handler",
    )

    def __init__(
        self,
        api: API,
        *,
        timeout: int | None = None,
        limit: int | None = None,
        offset: int = DEFAULT_OFFSET,
        reconnect_after: float = DEFAULT_RECONNECT_AFTER,
        max_reconnects: int = DEFAULT_MAX_RECONNECTS,
        include_updates: set[UpdateType] | None = None,
        exclude_updates: set[UpdateType] | None = None,
    ) -> None:
        self.api = api
        self.timeout = timeout
        self.limit = limit
        self.allowed_updates = self.get_allowed_updates(
            include_updates=include_updates,
            exclude_updates=exclude_updates,
        )
        self.reconnect_after = _compute_number(DEFAULT_RECONNECT_AFTER, reconnect_after, 0.0)
        self.max_reconnects = _compute_number(DEFAULT_MAX_RECONNECTS, max_reconnects, 0)
        self.offset = max(DEFAULT_OFFSET, offset)
        self._running = False
        self._reconnects_counter = 0
        self._error_handler = PollingErrorHandler(self)

    def __repr__(self) -> str:
        return (
            "<{}: api={!r}, running={}, offset={}, timeout={}, limit={}, "
            "allowed_updates={!r}, max_reconnects={}, reconnect_after={}>"
        ).format(
            type(self).__name__,
            self.api,
            self._running,
            self.offset,
            self.timeout,
            self.limit,
            self.allowed_updates,
            self.max_reconnects,
            self.reconnect_after,
        )

    @staticmethod
    def get_allowed_updates(
        *,
        include_updates: set[UpdateType] | None = None,
        exclude_updates: set[UpdateType] | None = None,
    ) -> list[UpdateType]:
        allowed_updates = list(UpdateType)

        if include_updates and exclude_updates:
            return [x for x in allowed_updates if x in include_updates and x not in exclude_updates]

        if exclude_updates:
            return [x for x in allowed_updates if x not in exclude_updates]

        if include_updates:
            return [x for x in allowed_updates if x in include_updates]

        return allowed_updates

    @property
    def reconnects_counter(self) -> int:
        return self._reconnects_counter

    @property
    def running(self) -> bool:
        return self._running

    def _reset_reconnects_counter(self) -> None:
        self._reconnects_counter = 0

    async def get_updates(self) -> msgspec.Raw:
        try:
            raw_updates = await self.api.request_raw(
                method="getUpdates",
                data=dict(
                    offset=self.offset,
                    limit=self.limit,
                    timeout=self.timeout,
                    allowed_updates=self.allowed_updates,
                ),
                timeout=self.timeout or 0 + self.api.http.timeout,
            )
        except TimeoutError:
            return msgspec.Raw(b"")

        if is_ok(raw_updates):
            return raw_updates.value

        match (error := raw_updates.error).status_code:
            case HTTPStatus.TOO_MANY_REQUESTS:
                error = APIServerError(
                    message="Too many requests to get updates",
                    retry_after=error.retry_after.unwrap_or(int(self.reconnect_after)),
                )
            case HTTPStatus.UNAUTHORIZED | HTTPStatus.NOT_FOUND as status:
                error = InvalidTokenError(
                    "Token seems to be invalid"
                    if status == HTTPStatus.NOT_FOUND
                    else "Invalid token (unauthorized)",
                )
            case HTTPStatus.BAD_GATEWAY | HTTPStatus.GATEWAY_TIMEOUT as status:
                error = APIServerError(
                    message="Telegram API server responded a {}".format(status.name.replace("_", " ")),
                    retry_after=int(self.reconnect_after),
                )

        raise error from None

    async def listen(self) -> typing.AsyncGenerator[list[Update], None]:
        logger.debug("Listening polling")
        self._running = True

        with decoder(list[Update]) as dec:
            while self._running:
                try:
                    if (raw := await self.get_updates()) and (updates := dec.decode(raw)):
                        yield updates
                        self.offset = updates[-1].update_id + 1

                    if self._reconnects_counter != 0:
                        self._reset_reconnects_counter()
                except BaseException as error:
                    if not await self._error_handler.handle(error):
                        logger.exception("Traceback message below:")

                    if isinstance(error, self.api.http.CONNECTION_TIMEOUT_ERRORS):
                        self._reconnects_counter += 1

    def stop(self) -> None:
        self._running = False
        self._reset_reconnects_counter()


__all__ = ("Polling",)



"""Code in this file is automatically parsed.
---
Nicifications are basically nice features for models which are included in auto-generate_noded models
The difference between nicifications and cure types is: cute types can borrow view runtime properties and have context api
(so they can implement model-specific methods).
Nicifications can only implement fields/methods/properties working only with model fields.
"""

from __future__ import annotations

from datetime import datetime
from functools import cached_property

from fntypes.option import Nothing, Option

from telegrinder.model import Model
from telegrinder.types import (
    Birthdate,
    Chat,
    ChatJoinRequest,
    ChatMemberUpdated,
    ChatType,
    ContentType,
    DefaultAccentColor,
    Message,
    ReplyKeyboardMarkup,
    ReplyKeyboardRemove,
    Update,
    UpdateType,
    User,
)


class _Birthdate(Birthdate):
    @property
    def is_birthday(self) -> bool:
        """True, if today is a user's birthday."""
        now = datetime.now()
        return now.month == self.month and now.day == self.day

    @property
    def age(self) -> Option[int]:
        """Optional. Contains the user's age, if the user has a birth year specified."""
        return self.year.map(lambda year: ((datetime.now() - datetime(year, self.month, self.day)) // 365).days)


class _Chat(Chat):
    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.id == other.id

    @property
    def full_name(self) -> Option[str]:
        """Optional. Full name (`first_name` + `last_name`) of the
        other party in a `private` chat.
        """
        return self.first_name.map(lambda x: x + " " + self.last_name.unwrap_or(""))


class _ChatJoinRequest(ChatJoinRequest):
    @property
    def chat_id(self) -> int:
        """`chat_id` instead of `chat.id`."""
        return self.chat.id


class _ChatMemberUpdated(ChatMemberUpdated):
    @property
    def chat_id(self) -> int:
        """Alias `.chat_id` instead of `.chat.id`"""
        return self.chat.id


class _Message(Message):
    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.message_id == other.message_id and self.chat_id == other.chat_id

    @cached_property
    def content_type(self) -> ContentType:
        """Type of content that the message contains."""
        for content in ContentType:
            if not isinstance(getattr(self, content.value, Nothing()), Nothing):
                return content

        return ContentType.UNKNOWN

    @property
    def from_user(self) -> "User":
        """`from_user` instead of `from_.unwrap()`."""
        return self.from_.unwrap()

    @property
    def chat_id(self) -> int:
        """`chat_id` instead of `chat.id`."""
        return self.chat.id

    @property
    def chat_title(self) -> str:
        """Chat title, for `supergroups`, `channels` and `group` chats.
        Full name, for `private` chat.
        """
        return self.chat.full_name.unwrap() if self.chat.type == ChatType.PRIVATE else self.chat.title.unwrap()


class _User(User):
    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.id == other.id

    @property
    def default_accent_color(self) -> DefaultAccentColor:
        """User's or bot's accent color (non-premium)."""
        return DefaultAccentColor(self.id % 7)

    @property
    def full_name(self) -> str:
        """User's or bot's full name (`first_name` + `last_name`)."""
        return self.first_name + self.last_name.map(lambda v: " " + v).unwrap_or("")


class _Update(Update):
    def __eq__(self, other: object, /) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.update_type == other.update_type

    @cached_property
    def update_type(self) -> UpdateType:
        """Incoming update type."""
        return UpdateType(next(iter(self.to_dict(exclude_fields={"update_id"}))))

    @cached_property
    def incoming_update(self) -> Model:
        """Incoming update."""
        return getattr(self, self.update_type.value).unwrap()


class _ReplyKeyboardMarkup(ReplyKeyboardMarkup):
    @property
    def empty_markup(self) -> ReplyKeyboardRemove:
        """Empty keyboard to remove the custom keyboard."""
        return ReplyKeyboardRemove(remove_keyboard=True, selective=self.selective.unwrap_or_none())



from __future__ import annotations

import abc
import typing
from functools import cached_property

import msgspec
from fntypes.result import Result

if typing.TYPE_CHECKING:
    from dataclasses import Field

    from _typeshed import DataclassInstance

type ModelType = DataclassWithIdentKey | ModelWithIdentKey | msgspec.Struct | DataclassInstance


@typing.runtime_checkable
class DataclassWithIdentKey(typing.Protocol):
    __key__: str
    __dataclass_fields__: typing.ClassVar[dict[str, Field[typing.Any]]]


@typing.runtime_checkable
class ModelWithIdentKey(typing.Protocol):
    __key__: str
    __struct_fields__: typing.ClassVar[tuple[str, ...]]
    __struct_config__: typing.ClassVar[msgspec.structs.StructConfig]


class ABCDataSerializer[Data](abc.ABC):
    ident_key: str | None = None

    @abc.abstractmethod
    def __init__(self, data_type: type[Data], /) -> None:
        pass

    @cached_property
    def key(self) -> str:
        return self.ident_key + "_" if self.ident_key else ""

    @abc.abstractmethod
    def serialize(self, data: Data) -> str:
        pass

    @abc.abstractmethod
    def deserialize(self, serialized_data: str) -> Result[Data, str]:
        pass


__all__ = ("ABCDataSerializer",)



import pathlib
import secrets
import typing

from telegrinder.msgspec_utils import encoder

type Files = dict[str, tuple[str, bytes]]


class InputFile:
    """Object `InputFile`, see the [documentation](https://core.telegram.org/bots/api#inputfile).

    This object represents the contents of a file to be uploaded. Must be posted using `multipart/form-data` in the usual way that files are uploaded via the browser.
    """

    __slots__ = ("filename", "data")

    filename: str
    """File name."""

    data: bytes
    """Bytes of file."""

    def __init__(self, filename: str, data: bytes) -> None:
        self.filename = filename
        self.data = data

    def __repr__(self) -> str:
        return "{}(filename={!r}, data={!r})".format(
            type(self).__name__,
            self.filename,
            (self.data[:30] + b"...") if len(self.data) > 30 else self.data,
        )

    @classmethod
    def from_path(cls, path: str | pathlib.Path, /) -> typing.Self:
        path = pathlib.Path(path)
        return cls(path.name, path.read_bytes())

    def _to_multipart(self, files: Files, /) -> str:
        attach_name = secrets.token_urlsafe(16)
        files[attach_name] = (self.filename, self.data)
        return f"attach://{attach_name}"


@encoder.add_enc_hook(InputFile)
def encode_input_file(input_file: InputFile, files: Files) -> str:
    return input_file._to_multipart(files)


__all__ = ("InputFile",)



from __future__ import annotations

import abc
import inspect
from collections import deque
from functools import reduce
from itertools import islice
from types import CodeType, NoneType, UnionType, resolve_bases

import typing_extensions as typing

from telegrinder.node.context import NODE_CONTEXT
from telegrinder.node.exceptions import ComposeError
from telegrinder.node.scope import NodeScope
from telegrinder.node.session import NodeSession
from telegrinder.tools.aio import Generator
from telegrinder.tools.fullname import fullname
from telegrinder.tools.magic.function import function_context, get_func_annotations
from telegrinder.tools.strings import to_pascal_case

if typing.TYPE_CHECKING:
    from telegrinder.node.tools.generator import generate_node
else:

    def generate_node(*args, **kwargs):
        from telegrinder.node.tools.generator import generate_node

        return generate_node(*args, **kwargs)


type NodeType = Node | NodeProto[typing.Any]
type IsNode = NodeType | type[NodeType]
type AnyNode = IsNode | NodeConvertable

T = typing.TypeVar("T", default=typing.Any)

ComposeResult: typing.TypeAlias = T | typing.Awaitable[T] | Generator[T, typing.Any, typing.Any]

_NODEFAULT: typing.Final[object] = object()
_NONE_TYPES: typing.Final[set[typing.Any]] = {None, NoneType}
_UNION_TYPES: typing.Final[set[typing.Any]] = {typing.Union, UnionType}
UNWRAPPED_NODE_KEY: typing.Final[str] = "__unwrapped_node__"


def is_node(maybe_node: typing.Any, /) -> typing.TypeIs[AnyNode]:
    return hasattr(maybe_node, "as_node") or is_node_type(maybe_node)


@typing.overload
def as_node(maybe_node: typing.Any, /) -> IsNode: ...


@typing.overload
def as_node(
    maybe_node: typing.Any,
    /,
    *,
    raise_exception: typing.Literal[False],
) -> IsNode | None: ...


def union_as_node(union: UnionType, /) -> IsNode | None:
    from telegrinder.node.either import _Either

    args = typing.get_args(union)
    if not args:
        return None

    plain, opt = [t for t in args if t not in _NONE_TYPES], any(t in _NONE_TYPES for t in args)
    if not plain:
        return None

    nodes = typing.cast("IsNode | tuple[IsNode, ...] | None", as_node(*plain, raise_exception=False))
    if nodes is None:
        return None

    nodes = (nodes,) if not isinstance(nodes, tuple) else nodes
    node = reduce(
        lambda left, right: _Either[left, right],
        islice(nodes, 1, None),
        nodes[0],  # type: ignore
    )
    return _Either[node] if opt else node


@typing.overload
def as_node(*maybe_nodes: typing.Any) -> tuple[IsNode, ...]: ...


@typing.overload
def as_node(
    *maybe_nodes: typing.Any,
    raise_exception: typing.Literal[False],
) -> tuple[IsNode, ...] | None: ...


def as_node(
    *maybe_nodes: typing.Any,
    raise_exception: bool = True,
) -> IsNode | tuple[IsNode, ...] | None:
    nodes = []

    for maybe_node in maybe_nodes:
        if isinstance(maybe_node, typing.TypeAliasType):
            maybe_node = maybe_node.__value__

        if (typing.get_origin(maybe_node) or maybe_node) in _UNION_TYPES and (
            maybe_node := union_as_node(union := maybe_node)
        ) is None:
            if not raise_exception:
                return None
            raise TypeError(f"Union `{union!r}` doesn't contain all types of Node.")

        if not is_node_type(orig := typing.get_origin(maybe_node) or maybe_node):
            if not hasattr(orig, "as_node"):
                if not raise_exception:
                    return None

                raise TypeError(
                    f"{'Type of' if isinstance(maybe_node, type) else 'Object of type'} "
                    f"{fullname(maybe_node)!r} cannot be resolved as Node.",
                )

            maybe_node = orig.as_node()

        nodes.append(maybe_node)

    return nodes[0] if len(nodes) == 1 else tuple(nodes)


def is_node_type(obj: typing.Any, /) -> typing.TypeIs[IsNode]:
    return isinstance(obj, Node | NodeProto) or (isinstance(obj, type) and issubclass(obj, Node | NodeProto))


@function_context("nodes")
def get_nodes(
    function: typing.Callable[..., typing.Any],
    /,
    *,
    start_idx: int = 0,
) -> dict[str, IsNode]:
    return {
        k: node
        for index, (k, v) in enumerate(get_func_annotations(function).items())
        if (node := as_node(v, raise_exception=False)) is not None and index >= start_idx
    }


@function_context("is_generator")
def is_generator(
    function: typing.Callable[..., typing.Any],
    /,
) -> typing.TypeGuard[Generator[typing.Any, typing.Any, typing.Any]]:
    return inspect.isgeneratorfunction(function) or inspect.isasyncgenfunction(function)


def unwrap_node(node: IsNode, /) -> tuple[IsNode, ...]:
    """Unwrap node as flattened tuple of node types in ordering required to calculate given node.

    Provides caching for passed node type.
    """
    if (unwrapped := getattr(node, UNWRAPPED_NODE_KEY, None)) is not None:
        return unwrapped

    stack = deque([(node, node.get_subnodes().values(), [node])])
    visited = list[IsNode]()

    while stack:
        parent, child_nodes, path = stack.pop()
        dependencies = set(child_nodes)

        if parent in dependencies:
            raise ComposeError(f"Node `{fullname(parent)}` refers to itself in dependency tree.")

        if parent not in visited:
            visited.insert(0, parent)

        for child in child_nodes:
            subnodes = child.get_subnodes().values()
            dependencies.update(subnodes)
            if child in path:
                raise ComposeError(
                    f"Cannot resolve node `{fullname(node)}` due to circular dependency "
                    f"({' -> '.join(fullname(n) for n in path[path.index(child) :] + [child])} <...>)",
                )

            stack.append((child, subnodes, path + [child]))

    unwrapped = tuple(visited)
    setattr(node, UNWRAPPED_NODE_KEY, unwrapped)
    return unwrapped


@typing.runtime_checkable
class Composable[R](typing.Protocol):
    @classmethod
    def compose(cls, *args: typing.Any, **kwargs: typing.Any) -> ComposeResult[R]: ...


class NodeConvertable(typing.Protocol):
    @classmethod
    def as_node(cls) -> type[NodeProto[typing.Any]]: ...


class NodeComposeFunction[R](typing.Protocol):
    __name__: str
    __code__: CodeType

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> ComposeResult[R]: ...


@typing.runtime_checkable
class NodeProto[R](Composable[R], typing.Protocol):
    @classmethod
    def get_subnodes(cls) -> dict[str, IsNode]: ...

    @classmethod
    def is_generator(cls) -> bool: ...


class Node(abc.ABC):
    node: str = "node"
    scope: NodeScope = NodeScope.PER_EVENT

    @classmethod
    @abc.abstractmethod
    def compose(cls, *args, **kwargs) -> ComposeResult:
        pass

    @classmethod
    def get_subnodes(cls) -> dict[str, IsNode]:
        return get_nodes(cls.compose)

    @classmethod
    def is_generator(cls) -> bool:
        return is_generator(cls.compose)


class scalar_node[T]:  # noqa: N801
    @typing.overload
    def __new__(cls, x: NodeComposeFunction[Composable[T]], /) -> type[T]: ...

    @typing.overload
    def __new__(cls, x: NodeComposeFunction[T], /) -> type[T]: ...

    @typing.overload
    def __new__(
        cls,
        /,
        *,
        scope: NodeScope,
    ) -> typing.Callable[[NodeComposeFunction[Composable[T]]], type[T]]: ...

    @typing.overload
    def __new__(
        cls,
        /,
        *,
        scope: NodeScope,
    ) -> typing.Callable[[NodeComposeFunction[T]], type[T]]: ...

    def __new__(cls, x=None, /, *, scope=_NODEFAULT) -> typing.Any:
        def inner(node_or_func, /) -> typing.Any:
            namespace = {"node": "scalar", "__module__": node_or_func.__module__}

            if isinstance(node_or_func, type):
                bases: list[type[typing.Any]] = [node_or_func]
                node_bases = resolve_bases(node_or_func.__bases__)

                if not any(issubclass(base, Node) for base in node_bases if isinstance(base, type)):
                    bases.append(Node)
                    namespace["scope"] = NodeScope.PER_EVENT if scope is _NODEFAULT else scope
                elif scope is not _NODEFAULT:
                    namespace["scope"] = scope

                return type(node_or_func.__name__, tuple(bases), namespace)
            else:
                base_node = generate_node(
                    func=node_or_func,
                    subnodes=tuple(get_nodes(node_or_func).values()),
                )
                return type(to_pascal_case(node_or_func.__name__), (base_node,), namespace)

        return inner if x is None else inner(x)


@typing.dataclass_transform(kw_only_default=True)
class FactoryNode(Node, abc.ABC):
    node = "factory"

    @classmethod
    @abc.abstractmethod
    def compose(cls, *args, **kwargs) -> ComposeResult:
        pass

    def __new__(cls, **context: typing.Any) -> type[typing.Self]:
        namespace = dict(**cls.__dict__)
        namespace.pop("__new__", None)
        return type(cls.__name__, (cls,), namespace | context)  # type: ignore


@typing.dataclass_transform()
class DataNode(Node, abc.ABC):
    node = "data"

    @classmethod
    @abc.abstractmethod
    def compose(cls, *args, **kwargs) -> ComposeResult[typing.Self]:
        pass


@scalar_node(scope=NodeScope.PER_CALL)
class Name:
    @classmethod
    def compose(cls) -> str: ...


@scalar_node
class NodeClass:
    @classmethod
    def compose(cls) -> type[Node]: ...


class GlobalNode[Value](Node):
    scope = NodeScope.GLOBAL

    @classmethod
    def set(cls, value: Value, /) -> None:
        NODE_CONTEXT.global_session[cls] = NodeSession(cls, value)

    @typing.overload
    @classmethod
    def get(cls) -> Value: ...

    @typing.overload
    @classmethod
    def get[Default](cls, *, default: Default) -> Value | Default: ...

    @classmethod
    def get(cls, **kwargs: typing.Any) -> typing.Any:
        sentinel = object()
        default = kwargs.pop("default", sentinel)

        if default is not sentinel and cls not in NODE_CONTEXT.global_sessions:
            return default

        if (session := NODE_CONTEXT.global_sessions.get(cls)) is None and default is sentinel:
            raise ValueError(f"Node `{fullname(cls)}` has no global value.")

        return session.value if session is not None else default

    @classmethod
    def unset(cls) -> None:
        NODE_CONTEXT.global_sessions.pop(cls, None)


__all__ = (
    "Composable",
    "ComposeError",
    "DataNode",
    "FactoryNode",
    "GlobalNode",
    "IsNode",
    "Name",
    "Node",
    "NodeClass",
    "NodeConvertable",
    "NodeProto",
    "NodeType",
    "as_node",
    "get_nodes",
    "is_node",
    "is_node_type",
    "scalar_node",
    "unwrap_node",
)



from __future__ import annotations

import abc
import typing

from telegrinder.tools.fullname import fullname
from telegrinder.tools.keyboard.buttons.base import BaseButton, BaseStaticButton
from telegrinder.tools.keyboard.buttons.tools import RowButtons

if typing.TYPE_CHECKING:
    from telegrinder.types.objects import InlineKeyboardMarkup, ReplyKeyboardMarkup

type DictStrAny = dict[str, typing.Any]
type AnyMarkup = ReplyKeyboardMarkup | InlineKeyboardMarkup
type RawKeyboard = list[list[DictStrAny]]
type Button = DictStrAny | list[DictStrAny] | BaseButton | RowButtons[BaseButton]

KEYBOARD_BUTTON_CLASS_KEY: typing.Final[str] = "_button_class"
NO_VALUE: typing.Final[object] = object()


def _has_under(name: str, /) -> bool:
    return name.startswith("_")


def _get_buttons(cls_: type[ABCStaticKeyboard], /) -> dict[str, BaseStaticButton]:
    button_class = cls_.get_button_class()
    return {k: v for k, v in dict(vars(cls_)).items() if isinstance(v, button_class)}


def copy_keyboard(keyboard: RawKeyboard, /) -> RawKeyboard:
    return [row.copy() for row in keyboard if row]


class ABCKeyboard(abc.ABC):
    keyboard: RawKeyboard

    @abc.abstractmethod
    def dict(self) -> DictStrAny:
        pass

    @abc.abstractmethod
    def get_markup(self) -> AnyMarkup:
        pass

    @abc.abstractmethod
    def copy(self, **with_changes: typing.Any) -> typing.Self:
        pass

    def add(self, button: Button, /) -> typing.Self:
        if not self.keyboard:
            self.row()

        if isinstance(button, RowButtons):
            self.keyboard[-1].extend(button.get_data())
            if button.auto_row:
                self.row()
            return self

        if isinstance(button, list):
            self.keyboard[-1].extend(button)
            return self

        self.keyboard[-1].append(button if isinstance(button, dict) else button.get_data())
        return self

    def row(self) -> typing.Self:
        if len(self.keyboard) and not len(self.keyboard[-1]):
            return self

        self.keyboard.append([])
        return self

    def format_text(self, **format_data: typing.Any) -> typing.Self:
        copy_keyboard = self.copy()

        for row in self.keyboard:
            for button in row:
                button["text"] = button["text"].format(**format_data)

        return copy_keyboard

    def merge(self, other: typing.Self, /) -> typing.Self:
        self.keyboard.extend(copy_keyboard(other.keyboard))
        return self

    def merge_to_last_row(self, other: typing.Self, /) -> typing.Self:
        kb_len = len(other.keyboard)

        for index, row in enumerate(copy_keyboard(other.keyboard), start=1):
            for button in row:
                self.keyboard[-1].append(button)

            if index < kb_len:
                self.keyboard.append([])

        return self


class StaticKeyboardMeta(type):
    if not typing.TYPE_CHECKING:

        def __getattr__(cls, name, /):
            if not _has_under(name):
                buttons = _get_buttons(cls)
                if name in buttons:
                    return buttons[name]
            return super().__getattribute__(name)

    def __setattr__(cls, name: str, value: typing.Any, /) -> None:
        if not _has_under(name) and name in _get_buttons(cls):  # type: ignore
            raise AttributeError(f"Cannot reassing attribute {name!r}.")
        return super().__setattr__(name, value)

    def __delattr__(cls, name: str, /) -> None:
        if not _has_under(name) and name in _get_buttons(cls):  # type: ignore
            raise AttributeError(f"Cannot delete attribute {name!r}.")
        return super().__delattr__(name)


class ABCStaticKeyboardMeta(abc.ABCMeta, StaticKeyboardMeta):
    pass


class ABCStaticKeyboard(metaclass=ABCStaticKeyboardMeta):
    __keyboard__: typing.Any

    @abc.abstractmethod
    def dict(self) -> DictStrAny:
        pass

    @abc.abstractmethod
    def get_markup(self) -> AnyMarkup:
        pass

    @classmethod
    def get_button_class(cls) -> type[BaseStaticButton]:
        sentinel = object()
        button_class = getattr(cls, KEYBOARD_BUTTON_CLASS_KEY, sentinel)

        if button_class not in (sentinel, NO_VALUE):
            return typing.cast("type[BaseStaticButton]", button_class)

        if button_class is not NO_VALUE:
            for obj in cls.__dict__.values():
                if isinstance(obj, BaseStaticButton):
                    button_class = type(obj)
                    setattr(cls, KEYBOARD_BUTTON_CLASS_KEY, button_class)
                    return button_class

        setattr(cls, KEYBOARD_BUTTON_CLASS_KEY, NO_VALUE)
        raise ValueError(f"Static keyboard {fullname(cls)!r} has no static buttons.")

    @classmethod
    def get_buttons(cls) -> dict[str, BaseStaticButton]:
        return _get_buttons(cls)


__all__ = ("ABCKeyboard", "ABCStaticKeyboard")



import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.tools.aio import maybe_awaitable
from telegrinder.types.objects import Update


class FuncRule(ABCRule):
    def __init__(
        self,
        func: typing.Callable[[Update, Context], typing.Awaitable[bool] | bool],
        /,
    ) -> None:
        self.func = func

    async def check(self, update: Update, ctx: Context) -> bool:
        return await maybe_awaitable(self.func(update, ctx))


__all__ = ("FuncRule",)



import pathlib

from telegrinder import API, Message, Telegrinder, Token
from telegrinder.bot.rules.is_from import IsPrivate
from telegrinder.modules import logger
from telegrinder.rules import IsUser, Text
from telegrinder.types import InputFile

api = API(token=Token.from_env())
bot = Telegrinder(api)
cool_bytes = pathlib.Path("examples/assets/satie.jpeg").read_bytes()

logger.set_level("INFO")


@bot.on.message(Text("/photo"), IsPrivate() & IsUser())
async def start(message: Message):
    await message.answer_photo(
        InputFile("satie.jpeg", cool_bytes),
        caption="Erik",
    )


bot.run_forever()



from abc import ABC
from functools import cached_property

import typing_extensions as typing
from fntypes.result import Error, Ok

from telegrinder.api.api import API
from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.modules import logger
from telegrinder.node.base import IsNode, get_nodes
from telegrinder.node.composer import compose_nodes
from telegrinder.tools.aio import maybe_awaitable
from telegrinder.tools.lifespan import Lifespan
from telegrinder.tools.magic.function import bundle
from telegrinder.types.objects import Update

type Event = Update | BaseCute[typing.Any]
type MiddlewareResult = bool | None | typing.Coroutine[typing.Any, typing.Any, bool | None]


async def run_middleware(
    method: typing.Callable[..., MiddlewareResult],
    api: API,
    event: Update,
    ctx: Context,
    required_nodes: typing.Mapping[str, IsNode] | None = None,
) -> bool | None:
    node_col = None
    data = {API: api, Update: event, Context: ctx}

    if required_nodes:
        match await compose_nodes(required_nodes, ctx, data=data):
            case Ok(value):
                node_col = value
            case Error(compose_error):
                logger.debug(
                    "Cannot compose nodes for `{}`, error: {!r}",
                    method.__qualname__,
                    compose_error.message,
                )
                return False

    try:
        bundle_method = bundle(method, ctx | ({} if node_col is None else node_col.values))
        bundle_method &= bundle(method, data, typebundle=True)
        return await maybe_awaitable(bundle_method())
    finally:
        if node_col is not None:
            await node_col.close_all()


class ABCMiddleware(ABC):
    def __repr__(self) -> str:
        name = f"middleware {type(self).__name__!r}"
        middleware_class = type(self)

        if middleware_class.post is not ABCMiddleware.post:
            name = "post-" + name

        if middleware_class.pre is not ABCMiddleware.pre:
            name = "pre-" + name

        return f"<{name}>"

    if typing.TYPE_CHECKING:

        def pre(self, *args: typing.Any, **kwargs: typing.Any) -> MiddlewareResult: ...

        def post(self, *args: typing.Any, **kwargs: typing.Any) -> MiddlewareResult: ...

    else:

        def pre(self, *args, **kwargs): ...

        def post(self, *args, **kwargs): ...

    @cached_property
    def pre_required_nodes(self) -> dict[str, IsNode]:
        return get_nodes(self.pre)

    @cached_property
    def post_required_nodes(self) -> dict[str, IsNode]:
        return get_nodes(self.post)

    def to_lifespan(
        self,
        event: Event,
        ctx: Context,
        *,
        api: API | None = None,
    ) -> Lifespan:
        if isinstance(event, BaseCute):
            event, api = event.raw_update, event.api

        if api is None:
            raise LookupError("Cannot get api, please pass as kwarg or provide BaseCute api-bound event.")

        return Lifespan(
            startup_tasks=[
                run_middleware(
                    self.pre,
                    api,
                    event,
                    ctx,
                    required_nodes=self.pre_required_nodes,
                ),
            ],
            shutdown_tasks=[
                run_middleware(
                    self.post,
                    api,
                    event,
                    ctx,
                    required_nodes=self.post_required_nodes,
                ),
            ],
        )


__all__ = ("ABCMiddleware", "run_middleware")



import dataclasses
import typing

import msgspec
from fntypes.option import Some

from telegrinder.api.api import API
from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.msgspec_utils import decoder
from telegrinder.node.base import ComposeError, FactoryNode, scalar_node
from telegrinder.tools.fullname import fullname
from telegrinder.types.objects import Model, Update

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance

type DataclassType = DataclassInstance | msgspec.Struct | dict[str, typing.Any]


@scalar_node
class UpdateNode:
    @classmethod
    def compose(cls, update: Update, api: API, context: Context) -> UpdateCute:
        match context.update_cute:
            case Some(update_cute):
                return update_cute
            case _:
                return context.add_update_cute(update, api).update_cute.unwrap()


class _EventNode(FactoryNode):
    dataclass: type[DataclassType]
    orig_dataclass: type[DataclassType]

    def __class_getitem__(cls, dataclass: type[DataclassType], /) -> typing.Self:
        return cls(dataclass=dataclass, orig_dataclass=typing.get_origin(dataclass) or dataclass)

    @classmethod
    def compose(cls, update_cute: UpdateNode, raw_update: Update, api: API) -> DataclassType:
        if cls.orig_dataclass is UpdateCute:
            return update_cute

        if issubclass(cls.orig_dataclass, BaseCute | Model):
            incoming_update = (
                update_cute.incoming_update
                if issubclass(cls.orig_dataclass, BaseCute)
                else raw_update.incoming_update
            )

            if type(incoming_update) is not cls.orig_dataclass:
                raise ComposeError(f"Incoming update is not `{fullname(cls.orig_dataclass)}`.")

            return incoming_update

        try:
            if issubclass(cls.orig_dataclass, msgspec.Struct) or dataclasses.is_dataclass(
                cls.orig_dataclass,
            ):
                dataclass = decoder.convert(
                    obj=raw_update.incoming_update,
                    type=cls.dataclass,
                    from_attributes=True,
                )
            else:
                dataclass = cls.dataclass(**raw_update.incoming_update.to_full_dict())

            return dataclass
        except Exception as exc:
            raise ComposeError(f"Cannot parse an update object into `{fullname(cls.dataclass)}`, error: {exc}")


if typing.TYPE_CHECKING:
    type EventNode[Dataclass: DataclassType] = Dataclass
else:
    EventNode = type("EventNode", (_EventNode,), {"__module__": __name__})


__all__ = ("EventNode",)



import typing
from abc import ABC, abstractmethod

from fntypes.result import Result

from telegrinder.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.types.objects import Update


class ABCHandler(ABC):
    final: bool

    @abstractmethod
    async def run(
        self,
        api: API,
        event: Update,
        context: Context,
        check: bool = True,
    ) -> Result[typing.Any, typing.Any]:
        pass


__all__ = ("ABCHandler",)



import dataclasses

from telegrinder import (
    API,
    CallbackQuery,
    InlineButton,
    InlineKeyboard,
    Message,
    Telegrinder,
    Token,
)
from telegrinder.modules import logger
from telegrinder.node.payload import PayloadData
from telegrinder.rules import (
    PayloadEqRule,
    PayloadMarkupRule,
    PayloadModelRule,
    Text,
)
from telegrinder.tools.callback_data_serialization import MsgPackSerializer

logger.set_level("DEBUG")
api = API(token=Token.from_env())
bot = Telegrinder(api)


@dataclasses.dataclass(slots=True, frozen=True)
class Item:
    __key__ = "item"

    name: str
    amount: int = dataclasses.field(kw_only=True)


item_serializer = MsgPackSerializer(Item)
kb = (
    InlineKeyboard()
    .add(InlineButton("Confirm", callback_data="confirm/action"))
    .row()
    .add(InlineButton("One", callback_data="number/1"))
    .add(InlineButton("Two", callback_data="number/2"))
    .row()
    .add(InlineButton("🍎", callback_data=Item("apple", amount=10), callback_data_serializer=item_serializer))
    .add(
        InlineButton("🍌", callback_data=Item("banana", amount=20), callback_data_serializer=item_serializer),
    )
    .row()
    .add(InlineButton("Won't respond", callback_data="number/foobar"))
).get_markup()


@bot.on.message(Text("/action"))
async def action(m: Message) -> None:
    await m.answer("Please confirm doing action.", reply_markup=kb)


@bot.on.callback_query(final=False)
async def handle_fruit_item(item: PayloadData[Item, MsgPackSerializer[Item]]) -> None:
    logger.info("Got fruit item={!r}", item)


@bot.on.callback_query(PayloadEqRule("confirm/action"))
async def callback_confirm_handler(cb: CallbackQuery) -> None:
    await cb.answer("Okay! Confirmed.")
    await cb.edit_text(text="Action happens.")
    # *some action happens*


@bot.on.callback_query(PayloadMarkupRule("number/<n:int>"))
async def callback_number_handler(cb: CallbackQuery, n: int) -> None:
    await cb.answer("{0} + (7 * 6) - {0} = 42🤯🤯🤯".format(n))


@bot.on.callback_query(PayloadModelRule(Item, serializer=MsgPackSerializer, alias="item"))
async def select_item(cb: CallbackQuery, item: Item) -> None:
    await cb.answer(f"You ate an {item.name!r} for {item.amount} cents 😋")


bot.run_forever()



from __future__ import annotations

import dataclasses
import typing

import msgspec
from fntypes.result import Result

from telegrinder.msgspec_utils.decoder import decoder
from telegrinder.msgspec_utils.encoder import encoder

if typing.TYPE_CHECKING:
    from telegrinder.api.error import APIError


def full_result[T](
    result: Result[msgspec.Raw, APIError],
    full_t: type[T],
) -> Result[T, APIError]:
    return result.map(lambda v: decoder.decode(v, type=full_t))


def get_params(params: dict[str, typing.Any], /) -> dict[str, typing.Any]:
    return {
        key: encoder.cast(value)
        for key, val in (
            *params.pop("other", {}).items(),
            *params.items(),
        )
        if key != "self" and (value := val.get() if isinstance(val, Proxy) else val) is not None
    }


@dataclasses.dataclass(frozen=True)
class Proxy[T]:
    cfg: ProxiedDict[T]
    key: str

    def get(self) -> typing.Any | None:
        return self.cfg._defaults.get(self.key)


class ProxiedDict[T]:
    if typing.TYPE_CHECKING:

        def __new__(cls, tp: type[T], /) -> T | typing.Self: ...

    def __init__(self, tp: type[T], /) -> None:
        self.tp = tp
        self._defaults: dict[str, typing.Any] = {}

    def __setattribute__(self, name: str, value: typing.Any, /) -> None:
        self._defaults[name] = value

    def __getitem__(self, key: str, /) -> typing.Any:
        return Proxy(self, key)

    def __setitem__(self, key: str, value: typing.Any, /) -> None:
        self._defaults[key] = value


__all__ = ("ProxiedDict", "full_result", "get_params")



from __future__ import annotations

import typing_extensions as typing

from telegrinder.api.api import API
from telegrinder.bot.dispatch import dispatch as dp
from telegrinder.bot.dispatch.abc import ABCDispatch
from telegrinder.bot.polling import polling as pg
from telegrinder.bot.polling.abc import ABCPolling
from telegrinder.modules import logger
from telegrinder.tools.global_context.builtin_context import TelegrinderContext
from telegrinder.tools.loop_wrapper import LoopWrapper

if typing.TYPE_CHECKING:
    from telegrinder.node.composer import Composer

Dispatch = typing.TypeVar("Dispatch", bound=ABCDispatch, default=dp.Dispatch)
Polling = typing.TypeVar("Polling", bound=ABCPolling, default=pg.Polling)

CONTEXT: typing.Final[TelegrinderContext] = TelegrinderContext()


class Telegrinder(typing.Generic[Dispatch, Polling]):
    def __init__(
        self,
        api: API,
        *,
        dispatch: Dispatch | None = None,
        polling: Polling | None = None,
        loop_wrapper: LoopWrapper | None = None,
    ) -> None:
        self.api = api
        self.dispatch = typing.cast("Dispatch", dispatch or dp.Dispatch())
        self.polling = typing.cast("Polling", polling or pg.Polling(api))
        self.loop_wrapper = loop_wrapper or CONTEXT.loop_wrapper

    def __repr__(self) -> str:
        return "<{}: api={!r}, dispatch={!r}, polling={!r}, loop_wrapper={!r}>".format(
            type(self).__name__,
            self.api,
            self.dispatch,
            self.polling,
            self.loop_wrapper,
        )

    @property
    def on(self) -> Dispatch:
        return self.dispatch

    @property
    def composer(self) -> Composer:
        return CONTEXT.composer.unwrap()

    async def reset_webhook(self) -> None:
        if not (await self.api.get_webhook_info()).unwrap().url:
            return
        await self.api.delete_webhook()

    async def run_polling(
        self,
        *,
        offset: int = 0,
        skip_updates: bool = False,
    ) -> typing.NoReturn:
        async def polling() -> typing.NoReturn:  # type: ignore
            if skip_updates:
                logger.debug("Dropping pending updates")
                await self.reset_webhook()
                await self.api.delete_webhook(drop_pending_updates=True)

            async for updates in self.polling.listen():
                for update in updates:
                    await self.loop_wrapper.create_task(self.dispatch.feed(update, self.api))

        self.polling.offset = offset

        if self.loop_wrapper.running:
            await polling()
        else:
            self.loop_wrapper.add_task(polling())
            self.loop_wrapper.run()

    def run_forever(self, *, offset: int = 0, skip_updates: bool = False) -> typing.NoReturn:
        logger.info("Running blocking polling (id={})", self.api.id)
        self.loop_wrapper.add_task(self.run_polling(offset=offset, skip_updates=skip_updates))
        self.loop_wrapper.run()


__all__ = ("Telegrinder",)



import typing

from fntypes.result import Error, Ok

from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.msgspec_utils import convert
from telegrinder.node.base import ComposeError, FactoryNode, Name, scalar_node


@scalar_node
class CallbackQueryData:
    @classmethod
    def compose(cls, callback_query: CallbackQueryCute) -> str:
        return callback_query.data.expect(ComposeError("Cannot complete decode callback query data."))


@scalar_node
class CallbackQueryDataJson:
    @classmethod
    def compose(cls, callback_query: CallbackQueryCute) -> dict:
        return callback_query.decode_data().expect(
            ComposeError("Cannot complete decode callback query data."),
        )


class _Field(FactoryNode):
    field_type: typing.Any

    def __class_getitem__(cls, field_type: typing.Any, /) -> typing.Self:
        return cls(field_type=field_type)

    @classmethod
    def compose(cls, callback_query_data: CallbackQueryDataJson, data_name: Name) -> typing.Any:
        if data := callback_query_data.get(data_name):
            match convert(data, cls.field_type):
                case Ok(value):
                    return value
                case Error(err):
                    raise ComposeError(err)

        raise ComposeError(f"Cannot find callback data with name {data_name!r}.")


if typing.TYPE_CHECKING:
    type Field[FieldType] = FieldType
else:
    Field = _Field


__all__ = (
    "CallbackQueryData",
    "CallbackQueryDataJson",
    "Field",
)



from telegrinder.bot.bot import Telegrinder
from telegrinder.bot.cute_types import (
    BaseCute,
    CallbackQueryCute,
    ChatJoinRequestCute,
    ChatMemberUpdatedCute,
    InlineQueryCute,
    MessageCute,
    PreCheckoutQueryCute,
    UpdateCute,
)
from telegrinder.bot.dispatch import (
    CALLBACK_QUERY_FOR_MESSAGE,
    CALLBACK_QUERY_FROM_CHAT,
    CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE,
    MESSAGE_FROM_USER,
    MESSAGE_FROM_USER_IN_CHAT,
    MESSAGE_IN_CHAT,
    ABCDispatch,
    ABCHandler,
    ABCMiddleware,
    ABCReturnManager,
    ABCView,
    AudioReplyHandler,
    BaseReturnManager,
    BaseView,
    CallbackQueryReturnManager,
    CallbackQueryView,
    ChatJoinRequestView,
    ChatMemberView,
    Context,
    Dispatch,
    DocumentReplyHandler,
    FuncHandler,
    Hasher,
    InlineQueryReturnManager,
    Manager,
    MediaGroupReplyHandler,
    MessageReplyHandler,
    MessageReturnManager,
    MessageView,
    PhotoReplyHandler,
    PreCheckoutQueryManager,
    PreCheckoutQueryView,
    RawEventView,
    ShortState,
    StickerReplyHandler,
    VideoReplyHandler,
    ViewBox,
    WaiterMachine,
    action,
    register_manager,
)
from telegrinder.bot.polling import ABCPolling, Polling
from telegrinder.bot.rules import ABCRule
from telegrinder.bot.scenario import ABCScenario, Checkbox, Choice

__all__ = (
    "CALLBACK_QUERY_FOR_MESSAGE",
    "CALLBACK_QUERY_FROM_CHAT",
    "CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE",
    "MESSAGE_FROM_USER",
    "MESSAGE_FROM_USER_IN_CHAT",
    "MESSAGE_IN_CHAT",
    "ABCDispatch",
    "ABCHandler",
    "ABCMiddleware",
    "ABCPolling",
    "ABCReturnManager",
    "ABCRule",
    "ABCScenario",
    "ABCView",
    "AudioReplyHandler",
    "BaseCute",
    "BaseReturnManager",
    "BaseView",
    "CallbackQueryCute",
    "CallbackQueryReturnManager",
    "CallbackQueryView",
    "ChatJoinRequestCute",
    "ChatJoinRequestView",
    "ChatMemberUpdatedCute",
    "ChatMemberView",
    "Checkbox",
    "Choice",
    "Context",
    "Dispatch",
    "DocumentReplyHandler",
    "FuncHandler",
    "Hasher",
    "InlineQueryCute",
    "InlineQueryReturnManager",
    "Manager",
    "MediaGroupReplyHandler",
    "MessageCute",
    "MessageReplyHandler",
    "MessageReturnManager",
    "MessageView",
    "PhotoReplyHandler",
    "Polling",
    "PreCheckoutQueryCute",
    "PreCheckoutQueryManager",
    "PreCheckoutQueryView",
    "RawEventView",
    "ShortState",
    "StickerReplyHandler",
    "Telegrinder",
    "UpdateCute",
    "VideoReplyHandler",
    "ViewBox",
    "WaiterMachine",
    "action",
    "register_manager",
)



"""The example uses the aiosqlite driver. It is recommended to use asynchronous drivers
to work with databases.
"""

import aiosqlite  # type: ignore
from fntypes.option import Nothing

from telegrinder import API, ABCMiddleware, Message, Telegrinder, Token
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.model import decoder
from telegrinder.modules import logger
from telegrinder.msgspec_utils import Option
from telegrinder.rules import IsReply, Markup, MessageEntities, Text
from telegrinder.types.enums import MessageEntityType
from telegrinder.types.objects import User

db_path = "examples/assets/users.db"
bot = Telegrinder(API(Token.from_env()))
logger.set_level("INFO")


@bot.loop_wrapper.lifespan.on_startup
async def create_table() -> None:
    async with aiosqlite.connect(db_path) as conn:  # noqa: SIM117
        async with conn.cursor() as cur:
            await cur.execute(
                "create table if not exists users("
                "id integer unique,"
                "is_bot boolean,"
                "is_premium boolean,"
                "first_name varchar(64),"
                "last_name varchar(64),"
                "username varchar(32),"
                "language_code varchar(5))"
            )
            await conn.commit()


def get_result_with_names(
    cursor: aiosqlite.Cursor,
    row: aiosqlite.Row,
    as_bool: set[str] | None = None,
) -> dict:
    column_names = [d[0] for d in cursor.description]
    resulting_row = {}
    for index, column_name in enumerate(column_names):
        value = row[index]
        resulting_row[column_name] = bool(value) if column_name in (as_bool or ()) and value in (0, 1) else value
    return resulting_row


class DummyDatabase:
    async def set_user(self, user: User) -> None:
        async with aiosqlite.connect(db_path) as conn:
            async with conn.cursor() as cur:
                await cur.execute(
                    """
                    INSERT OR REPLACE INTO users(
                        id, is_bot, first_name, last_name,
                        username, language_code, is_premium
                    ) VALUES (
                        :id, :is_bot, :first_name, :last_name,
                        :username, :language_code, :is_premium
                    );
                    """,
                    user.to_full_dict(),
                )
            await conn.commit()

    async def get_user_by_username(self, username: str) -> Option[User]:
        async with (
            aiosqlite.connect(db_path) as conn,
            conn.cursor() as cur,
        ):
            row = await (await cur.execute("SELECT * FROM users WHERE username = ?", (username,))).fetchone()
            if row is None:
                return Nothing()
            return decoder.convert(
                obj=get_result_with_names(cur, row, as_bool={"is_premium", "is_bot"}),
                type=Option[User],  # type: ignore
            )

    async def get_or_set_user(self, user: User) -> User:
        if (await self.get_user_by_id(user.id)).unwrap_or_none() is None:
            await self.set_user(user)
        return user

    async def get_user_by_id(self, user_id: int) -> Option[User]:
        async with (
            aiosqlite.connect(db_path) as conn,
            conn.cursor() as cur,
        ):
            row = await (await cur.execute("SELECT * FROM users WHERE id = ?", (user_id,))).fetchone()
            if row is None:
                return Nothing()
            return decoder.convert(
                obj=get_result_with_names(cur, row, as_bool={"is_premium", "is_bot"}),
                type=Option[User],  # type: ignore
            )


db = DummyDatabase()


@bot.on.message.register_middleware
class UserRegistrarMiddleware(ABCMiddleware):
    async def pre(self, event: Message) -> bool:
        if event.from_:
            await db.set_user(event.from_user)
        if event.reply_to_message and event.reply_to_message.unwrap().from_:
            await db.set_user(event.reply_to_message.unwrap().from_user)
        return True


class MentionRule(
    ABCRule,
    requires=[
        Text("/get_user") & IsReply()
        | Markup("/get_user id<user_id:int>")
        | MessageEntities(MessageEntityType.TEXT_MENTION) & Markup("/get_user <!>")
        | MessageEntities(MessageEntityType.MENTION) & Markup("/get_user @<username>")
    ],
):
    async def check(self, message: Message, ctx: Context) -> bool:
        user = None

        match ctx:
            case {"message_entities": message_entities} if (
                message_entities[0].type == MessageEntityType.TEXT_MENTION
            ):
                user = await db.get_or_set_user(message_entities[0].user.unwrap())
            case {"username": username}:
                user = (await db.get_user_by_username(username)).unwrap_or_none()
            case {"user_id": user_id}:
                user = (await db.get_user_by_id(user_id)).unwrap_or_none()
            case _ if message.reply_to_message:
                user = message.reply_to_message.unwrap().from_.unwrap_or_none()

        if user is None:
            return False

        ctx.set("mentioned_user", user)
        return True


@bot.on.message(MentionRule())
async def get_user(mentioned_user: User) -> str:
    return f"""
    Id -> {mentioned_user.id}
    Is bot -> {"✅" if mentioned_user.is_bot else "❌"}
    First name -> {mentioned_user.first_name}
    Last name -> {mentioned_user.last_name.unwrap_or(repr("UNKNOWN"))}
    Is premium -> {"✅" if mentioned_user.is_premium.unwrap_or(False) else "❌"}
    Language code -> {mentioned_user.language_code.unwrap_or(repr("UNKNOWN"))}
    """.replace("    ", "")


bot.run_forever()



import enum

from telegrinder import API, MemoryStateStorage, Message, StateData, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import StateMeta, Text

api = API(token=Token.from_env())
bot = Telegrinder(api)
states = MemoryStateStorage()

logger.set_level("INFO")


class StateEnum(enum.StrEnum):
    # You can't get blessed when cursed,
    # but you can get cursed when blessed...
    CURSED = "cursed"
    BLESSED = "blessed"


@bot.on.message(
    Text("/curse"),
    states.State(StateEnum.BLESSED) | states.State(StateMeta.NO_STATE),
)
async def curse_handler(m: Message):
    await states.set(m.from_user.id, StateEnum.CURSED, {})
    await m.answer("You are now cursed..")


@bot.on.message(
    Text("/bless"),
    states.State(StateMeta.NO_STATE),
)
async def bless_handler(m: Message):
    await states.set(m.from_user.id, StateEnum.BLESSED, {})
    await m.answer("YAS you are now blessed")


@bot.on.message()
async def any_message_handler(m: Message):
    state = await states.get(m.from_user.id)  # TODO: state node
    vibe = state.unwrap_or(StateData("normal", {})).key
    await m.answer(f"Omg you are so {vibe}..")


bot.run_forever()



from telegrinder.bot.dispatch.view.base import BaseView
from telegrinder.types.enums import UpdateType


class ChatJoinRequestView(BaseView):
    def __init__(self) -> None:
        super().__init__(UpdateType.CHAT_JOIN_REQUEST)


__all__ = ("ChatJoinRequestView",)



from telegrinder.tools.keyboard.base import ABCKeyboard, ABCStaticKeyboard, BaseKeyboard, BaseStaticKeyboard
from telegrinder.tools.keyboard.buttons.base import (
    BaseButton,
    BaseStaticButton,
)
from telegrinder.tools.keyboard.buttons.buttons import (
    Button,
    InlineButton,
)
from telegrinder.tools.keyboard.buttons.static_buttons import (
    StaticButton,
    StaticInlineButton,
)
from telegrinder.tools.keyboard.buttons.tools import RowButtons
from telegrinder.tools.keyboard.data import KeyboardModel
from telegrinder.tools.keyboard.keyboard import InlineKeyboard, Keyboard
from telegrinder.tools.keyboard.static_keyboard import (
    StaticInlineKeyboard,
    StaticKeyboard,
)

__all__ = (
    "ABCKeyboard",
    "ABCStaticKeyboard",
    "BaseButton",
    "BaseKeyboard",
    "BaseStaticButton",
    "BaseStaticKeyboard",
    "Button",
    "InlineButton",
    "InlineKeyboard",
    "Keyboard",
    "KeyboardModel",
    "RowButtons",
    "StaticButton",
    "StaticInlineButton",
    "StaticInlineKeyboard",
    "StaticKeyboard",
)



from .config import ConfigTOML, read_config
from .generator import (
    ABCGenerator,
    MethodGenerator,
    ObjectGenerator,
    generate,
)
from .merge_shortcuts import merge_shortcuts
from .models import (
    Config,
    MethodParameter,
    MethodSchema,
    Model,
    ObjectField,
    ObjectSchema,
    TelegramBotAPISchema,
    dec_hook,
)

__all__ = (
    "ABCGenerator",
    "Config",
    "ConfigTOML",
    "MethodGenerator",
    "MethodParameter",
    "MethodSchema",
    "Model",
    "ObjectField",
    "ObjectGenerator",
    "ObjectSchema",
    "TelegramBotAPISchema",
    "dec_hook",
    "generate",
    "merge_shortcuts",
    "read_config",
)



from telegrinder.bot.rules.abc import ABCRule
from telegrinder.bot.rules.node import NodeRule
from telegrinder.node.base import as_node
from telegrinder.node.text import TextInteger


class IsInteger(NodeRule):
    def __init__(self) -> None:
        super().__init__(as_node(TextInteger))


class IntegerInRange(ABCRule):
    def __init__(self, rng: range) -> None:
        self.rng = rng

    def check(self, integer: TextInteger) -> bool:
        return integer in self.rng


__all__ = ("IntegerInRange", "IsInteger")



from telegrinder.bot.dispatch.view.abc import ABCEventRawView
from telegrinder.bot.dispatch.view.base import BaseView


class RawEventView(ABCEventRawView, BaseView):
    def __init__(self) -> None:
        super().__init__()


__all__ = ("RawEventView",)



import typing

from telegrinder.node.base import ComposeError, IsNode, Node
from telegrinder.node.container import ContainerNode
from telegrinder.tools.aio import maybe_awaitable


def cast_false_to_none[Value](value: Value) -> Value | None:
    if value is False:
        return None
    return value


def error_on_none[Value](value: Value | None) -> Value:
    if value is None:
        raise ComposeError
    return value


def generate_node(
    subnodes: tuple[IsNode, ...],
    func: typing.Callable[..., typing.Any],
    casts: tuple[typing.Callable[[typing.Any], typing.Any], ...] = (cast_false_to_none, error_on_none),
) -> type[Node]:
    async def compose(_, *args: typing.Any) -> typing.Any:
        result = await maybe_awaitable(func(*args))
        for cast in casts:
            result = cast(result)
        return result

    return ContainerNode.link_nodes(linked_nodes=list(subnodes), composer=compose)


__all__ = ("generate_node",)



from examples.blueprint_bot.client import wm
from telegrinder import MESSAGE_FROM_USER, Dispatch, Message
from telegrinder.modules import logger
from telegrinder.rules import HasText, Text

dp = Dispatch()


@dp.message(Text("/funnel"))
async def funnel_handler(message: Message):
    await message.answer("What is your favourite colour of thursday evening")
    msg, _ = await wm.wait(MESSAGE_FROM_USER, message.from_user.id, release=HasText())
    await message.answer("Oki")
    logger.info(
        "{}'s favourite colour of thursday evening is {}",
        msg.from_user.first_name,
        msg.text,
    )
    await message.answer("And what is the pattern of the beautiful scarf you wear in the winter nights?")
    msg, _ = await wm.wait(MESSAGE_FROM_USER, message.from_user.id, release=HasText())
    logger.info(
        "{}'s pattern of the winter scarf is {}",
        msg.from_user.first_name,
        msg.text,
    )
    await message.answer("Brilliant")



import pathlib

import pytest

from .test_utils import MockedHttpClient

RESPONSE_KITTEN_BYTES = pathlib.Path("examples/assets/kitten.jpg").read_bytes()
RESPONSE_JSON = {
    "ok": True,
    "response": {
        "count": 1,
        "items": [
            {
                "item": "Box",
                "price": 5295.0,
                "quantity": 1,
                "description": "Box of things",
                "image": b"XXXXXXXXXXXXXXXXXX",
                "id": 9327156890021313285781,
                "created": "11.11.2009",
                "updated": "09.01.2019",
                "deleted": "30.12.2023",
                "from": "USA",
            }
        ],
    },
}


@pytest.mark.asyncio()
async def test_http_request_text():
    client = MockedHttpClient("response")
    response = await client.request_text("https://site.com")
    await client.close()

    assert isinstance(response, str)
    assert response.islower()
    assert response == "response"


@pytest.mark.asyncio()
async def test_http_request_content():
    client = MockedHttpClient(RESPONSE_KITTEN_BYTES)
    response = await client.request_content("https://kitten.com", data={"image": "jpg"})
    await client.close()

    assert isinstance(response, bytes)
    assert response == RESPONSE_KITTEN_BYTES


@pytest.mark.asyncio()
async def test_http_request_json():
    client = MockedHttpClient(RESPONSE_JSON)
    response = await client.request_json("https://megashop.net", data={"item": "Box"})
    await client.close()

    assert isinstance(response, dict)
    assert response == RESPONSE_JSON
    assert response["response"]["items"][0]["item"] == "Box"
    assert response["response"]["items"][0]["price"] == 5295.0


@pytest.mark.asyncio()
async def test_http_request_bytes():
    client = MockedHttpClient(b"response")
    response = await client.request_bytes("https://kitten.com", data={"image": "jpg"})
    await client.close()

    assert isinstance(response, bytes)
    assert response == b"response"



import types
import typing
from collections import OrderedDict
from datetime import timedelta
from urllib.parse import urlencode

from telegrinder.tools.magic.function import get_func_annotations

type DeepLinkFunction[**P] = typing.Callable[P, str]
type NoValue = types.EllipsisType

Parameter = typing.Annotated

NO_VALUE: typing.Final[NoValue] = typing.cast("NoValue", ...)


def get_query_params(
    func: DeepLinkFunction[...],
    kwargs: dict[str, typing.Any],
    order_params: set[str] | None = None,
) -> dict[str, typing.Any]:
    annotations = get_func_annotations(func)
    params = OrderedDict()
    param_names = (
        [*order_params, *(p for p in annotations if p not in order_params)] if order_params else annotations
    )

    for param_name in param_names:
        annotation = annotations[param_name]
        if param_name in kwargs:
            value = kwargs[param_name]
            if typing.get_origin(annotation) is Parameter:
                param_name, validator = get_parameter_metadata(annotation)
                value = validator(value) if validator is not None else value

            params[param_name] = value

    return params


def parse_query_params(
    params: dict[str, typing.Any],
    no_value_params: set[str] | None = None,
    /,
) -> tuple[set[str], dict[str, typing.Any]]:
    no_value_params = no_value_params or set()
    params_: dict[str, typing.Any] = {}

    for key, value in params.items():
        if value in (False, None):
            continue

        if value in (True, NO_VALUE):
            no_value_params.add(key)
            continue
        if isinstance(value, timedelta):
            value = int(value.total_seconds())

        params_[key] = value

    return (no_value_params, params_)


def get_parameter_metadata(
    parameter: typing.Any,
) -> tuple[str, typing.Callable[[typing.Any], typing.Any] | None]:
    meta: tuple[typing.Any, ...] = getattr(parameter, "__metadata__")
    return meta if len(meta) == 2 else (meta[0], None)


def parse_deep_link(
    *,
    link: str,
    params: dict[str, typing.Any],
    no_value_params: set[str] | None = None,
) -> str:
    no_value_params, params = parse_query_params(params, no_value_params)
    query = urlencode(params, encoding="UTF-8") + ("&" if no_value_params else "") + "&".join(no_value_params)
    return f"{link}?{query}"


__all__ = (
    "NO_VALUE",
    "NoValue",
    "Parameter",
    "get_parameter_metadata",
    "get_query_params",
    "parse_deep_link",
    "parse_query_params",
)



import abc
import importlib
import keyword
import os
import pathlib
import re
import sys
import typing

import msgspec
import requests

from .config import ConfigTOML, read_config
from .merge_shortcuts import merge_shortcuts
from .models import (
    Config,
    MethodParameter,
    MethodSchema,
    MethodsParamsLiteralTypesParam,
    ObjectField,
    ObjectSchema,
    ObjectsFieldsLiteralTypesField,
    ObjectsIdByDefaultField,
    TelegramBotAPISchema,
    TypedDefaultParameter,
    dec_hook,
)

try:
    from telegrinder.modules import logger
except ImportError:
    import logging

    logger = logging.getLogger("typegen")

TYPEGEN_DIR: typing.Final[pathlib.Path] = pathlib.Path(__file__).parent
TAB: typing.Final[str] = "    "
MAX_LENGTH_LINE_CHUNK: typing.Final[int] = 60
TYPES: typing.Final[dict[str, str]] = {
    "String": "str",
    "Integer": "int",
    "Float": "float",
    "Boolean": "bool",
    "Unixtime": "datetime",
    "Timestamp": "timedelta",
}
INPUTFILE_DOCSTRING: typing.Final[str] = "using multipart/form-data"


def download_schema(config_toml: ConfigTOML, config_model: Config, /) -> TelegramBotAPISchema:
    logger.debug(f"Downloading schema from {config_model.telegram_bot_api.schema_url!r}...")

    schema: dict[str, typing.Any] = msgspec.json.decode(
        requests.get(config_model.telegram_bot_api.schema_url).text
    )
    schema_version = float(schema["version"].split()[-1])
    schema.update(
        dict(
            methods=[d for d in schema["methods"].values()],
            types=[d for d in schema["types"].values()],
        ),
    )

    logger.debug(
        "Schema (version={}, release_date={!r}) successfully downloaded!",
        schema_version,
        schema["release_date"],
    )

    if schema_version > config_model.telegram_bot_api.version_number:
        logger.debug("New version of the schema, upgrade version of the telegram bot api in config.")
        config_toml.document["telegram-bot-api"]["version"] = f"v{schema_version}"  # type: ignore
        config_toml.save()

    return msgspec.convert(obj=schema, type=TelegramBotAPISchema)


def find_object_nicifications(
    object_name: str,
    nicification_path: pathlib.Path,
    /,
) -> tuple[str | None, list[str]]:
    pattern = r"class _" + object_name + r"\((?P<base>.+)\):\n((?:.|\n {4}|\n$)+)"
    nicifications_source = nicification_path.read_text(encoding="UTF-8")
    matches = list(re.finditer(pattern, nicifications_source, flags=re.MULTILINE))
    return (None, []) if not matches else (matches[0].group("base"), [match.group(2) for match in matches])


def chunks_str(s: str, sep: str = "\n"):
    chunked_string, line_length = "", 0

    for word in s.split():
        chunked_string += word + " "
        line_length += len(word)

        if line_length >= MAX_LENGTH_LINE_CHUNK:
            chunked_string = chunked_string.strip() + sep
            line_length = 0

    return chunked_string.rstrip()


def makesafe_name(s: str, /) -> str:
    return s + "_" if keyword.iskeyword(s) else s


def run_ruff_formatter(path: pathlib.Path, /) -> None:
    logger.debug("Run ruff formatter...")
    if os.system(f"ruff format {path}") != 0:
        logger.error("ruff formatter failed.")
    else:
        logger.info("ruff formatter successfully formatted files.")

    logger.debug("Run ruff-isort...")
    if os.system(f"ruff check {path} --select I --select F401 --fix") != 0:
        logger.error("ruff-isort failed.")
    else:
        logger.info("ruff-isort successfully sorted imports.")

    logger.debug("Run ruff-sortall...")
    if os.system(f"ruff check {path} --select RUF022 --fix") != 0:
        logger.error("ruff-sortall failed.")
    else:
        logger.info("ruff-sortall successfully sorted dunder alls.")


def to_optional(st: str) -> str:
    if '"' in st:
        return '"{} | None"'.format(st.replace('"', ""))
    return st + " | None"


def convert_to_python_type(
    tp: str,
    parent_types: dict[str, list[str]] | None = None,
    as_forward_ref: bool = False,
    as_union: bool = False,
) -> str:
    if tp.startswith("Array of"):
        return "list[{}]".format(
            " | ".join(
                convert_to_python_type(t, parent_types, as_forward_ref, as_union)
                for t in tp.removeprefix("Array of ").split(", ")
            )
        )

    if tp.startswith("typing.Literal"):
        return tp

    parent_types = parent_types or {}
    hint = '"{}"' if as_forward_ref else "{}"

    if tp in TYPES:
        return TYPES[tp]
    return (
        hint.format(tp)
        if tp not in parent_types
        else (
            "Variative[%s]" % ", ".join(hint.format(x) for x in parent_types[tp])
            if not as_union
            else hint.format(" | ".join(x for x in parent_types[tp]))
        )
    )


def camel_to_snake(s: str) -> str:
    return "".join("_" + x.lower() if x.isupper() else x for x in s)


def camel_to_pascal(s: str) -> str:
    return (s[0].upper() if s[0].islower() else s[0]) + s[1:]


def is_unixtime_type(name: str, types: list[str], description: str) -> bool:
    return "Integer" in types and (
        name == "date"
        or name.endswith("_date")
        or any(x in description.lower() for x in ("unix timestamp", "unix time"))
    )


def is_timestamp_type(name: str, types: list[str]) -> bool:
    return any(("Integer" in types, "Float" in types)) and (name == "timestamp" or name.endswith("_timestamp"))


class ABCGenerator(abc.ABC):
    @abc.abstractmethod
    def generate(self, path: pathlib.Path) -> None:
        pass


class ObjectGenerator(ABCGenerator):
    def __init__(
        self,
        objects: list[ObjectSchema],
        config: Config,
    ) -> None:
        self.objects = objects
        self.config = config
        self.parent_types: dict[str, list[str]] = {obj.name: obj.subtypes for obj in objects if obj.subtypes}

    def get_literal_types_field(self, object_name: str, field_name: str) -> ObjectsFieldsLiteralTypesField | None:
        for object_literal_types in self.config.generator.objects.fields_literal_types:
            if object_literal_types.object_name == object_name:
                for field in object_literal_types.fields:
                    if field_name == field.name:
                        return field

        return None

    def get_generation_id_by_default_field(
        self, object_name: str, field_name: str
    ) -> ObjectsIdByDefaultField | None:
        for id_by_default in self.config.generator.objects.id_by_default:
            if object_name == id_by_default.object_name:
                for field in id_by_default.fields:
                    if field_name == field.name:
                        return field

        return None

    def make_object_field(
        self,
        field: ObjectField,
        literal_types: ObjectsFieldsLiteralTypesField | None = None,
    ) -> str:
        code = makesafe_name(field.name) + ": "
        field_type = "typing.Any"
        field_value = (
            "field(default=..., converter={converter})" if not field.required else "field(converter={converter})"
        )

        if literal_types is not None and any((literal_types.enum, literal_types.literals)):
            literal_type_hint = literal_types.enum or "Literal[%s]" % ", ".join(
                f'"{x}"' if isinstance(x, str) else str(x) for x in literal_types.literals
            )
            if len(literal_types.literals) > 3:
                literal_type_hint = literal_type_hint.replace("]", ",]")

            field_type = (
                f"list[{literal_type_hint}]" if any("Array of" in x for x in field.types) else literal_type_hint
            )
            field_value = (
                "field(default={})".format(field.default)
                if field.required and field.default is not None
                else "field()"
                if field.required
                else f"field(default=..., converter=From[{field_type} | None])"
            )
        else:
            if is_timestamp_type(field.name, field.types):
                field_type = "timedelta"
                converter_type = field_type

                if "Integer" in field.types:
                    field.types.remove("Integer")
                    converter_type += "| int"
                if "Float" in field.types:
                    field.types.remove("Float")
                    converter_type += "| float"

                field_value = (
                    "field(default=..., converter=From[{} | None])"
                    if not field.required
                    else "field(converter=From[{}])"
                ).format(converter_type)

            elif is_unixtime_type(field.name, field.types, field.description or ""):
                field.types.remove("Integer")
                field_type = "datetime"
                field_value = (
                    "field(default=..., converter=From[datetime | int | None])"
                    if not field.required
                    else "field(converter=From[datetime | int])"
                )

            if "InputFile" in field.types:
                field.types.append(field.types.pop(field.types.index("InputFile")))

            if field.description and "InputFile" not in field.types and INPUTFILE_DOCSTRING in field.description:
                field.types.append("InputFile")

            if len(field.types) > 1:
                field_type = "Variative[%s]" % ", ".join(
                    convert_to_python_type(tp, self.parent_types) for tp in field.types
                )
                union_types = " | ".join(
                    convert_to_python_type(tp, self.parent_types, as_union=True, as_forward_ref=True)
                    for tp in field.types
                )
                if '"' in union_types:
                    union_types = '"{}"'.format(union_types.replace('"', ""))
                field_value = field_value.format(
                    converter="From[%s]" % (to_optional(union_types) if not field.required else union_types)
                )

            elif len(field.types) == 1:
                field_type = convert_to_python_type(field.types[0], self.parent_types)
                converted_type = convert_to_python_type(
                    field.types[0],
                    self.parent_types,
                    as_forward_ref=True,
                    as_union=True,
                )
                field_value = (
                    field_value.format(
                        converter="From[%s]"
                        % (to_optional(converted_type) if not field.required else converted_type),
                    )
                    if ("|" in converted_type or not field.required)
                    else "field()"
                )

        if not field.required:
            field_type = f"Option[{field_type}]"

        if field.default_factory is not None:
            field_value = field_value.replace(")", f"default_factory={field.default_factory},)")

        code += f"{field_type} = {field_value}"
        if field.description:
            description = field.description.replace('"', "`")
            sep = "\n" + TAB
            code += (
                f'{sep}"""{chunks_str(description, sep=sep)}{"." if not description.endswith(".") else ""}"""\n'
            )

        return code

    def make_subtype_of(self, subtype_of: list[str]) -> str:
        return "Model" if not subtype_of else ", ".join(subtype_of)

    def make_object(self, object_schema: ObjectSchema) -> str:
        object_name = camel_to_pascal(object_schema.name)

        if self.config.generator.nicifications_path is not None:
            base_object_name, nicifications = find_object_nicifications(
                object_name,
                self.config.generator.nicifications_path,
            )
        else:
            base_object_name, nicifications = None, []

        sybtypes_of = self.make_subtype_of(object_schema.subtype_of or [])
        description = (
            "Base object"
            if object_schema.subtypes
            else (
                "Object"
                if base_object_name is None or base_object_name == object_name
                else base_object_name.split(".")[-1] + " object"
            )
        ) + f" `{object_name}`, see the [documentation]({object_schema.href})."

        code = (
            f"class {camel_to_pascal(object_schema.name)}"
            f"({base_object_name if base_object_name and base_object_name != object_name else sybtypes_of}):\n{TAB}"
        )
        code += '"""%s\n\n%s\n"""' % (
            description,
            ("No description yet." if not object_schema.description else "\n".join(object_schema.description)),
        )
        code += f"\n"

        if not object_schema.fields and not nicifications:
            if (not object_schema.subtypes and base_object_name is None) or not object_schema.subtypes:
                logger.warning(
                    f"Object {object_name!r} has no fields or subtypes or nicifications (mark as empty object).",
                )

            code += TAB + "pass" if not object_schema.description else ""
            return code

        if object_schema.fields:
            for f in object_schema.fields:
                literal_types = self.get_literal_types_field(object_name, f.name)
                if literal_types is not None and literal_types.literals and f.required:
                    f.default = literal_types.default or (
                        f'"{literal_types.literals[0]}"'
                        if isinstance(literal_types.literals[0], str)
                        else str(literal_types.literals[0])
                    )
                elif literal_types is not None and literal_types.enum_default is not None and f.required:
                    f.default = literal_types.enum_default

                generation_id_by_default = self.get_generation_id_by_default_field(object_name, f.name)
                if generation_id_by_default is not None:
                    f.default_factory = "lambda: secrets.token_urlsafe({nbytes})".format(
                        nbytes=generation_id_by_default.nbytes,
                    )

            code += TAB
            for field in (
                *filter(
                    lambda f: f.default is None and f.default_factory is None and f.required,
                    object_schema.fields,
                ),
                *filter(lambda f: f.default is not None, object_schema.fields),
                *filter(lambda f: f.default_factory is not None, object_schema.fields),
                *filter(lambda f: not f.required, object_schema.fields),
            ):
                literal_types = self.get_literal_types_field(object_name, field.name)
                code += f"\n{TAB}" + self.make_object_field(field, literal_types)

        for nicification in nicifications:
            if object_schema.fields:
                for f, t, d in re.findall(r'(\b\w+\b):\s*(.+)\s*"""(.*?)"""', nicification, flags=re.DOTALL):
                    new_code = f'\n    {f}: {t.strip()}\n    """{d}"""'
                    code = re.sub(r"\n    " + f + r": .+((?:.|\n {4}|\n$)+)", new_code, code)
                    nicification = nicification.replace(new_code.strip(), "")

            code += nicification

        return code

    def generate(self, path: pathlib.Path) -> None:
        if not self.objects:
            logger.error("Objects is empty.")
            sys.exit(-1)

        logger.debug("Generate objects...")
        lines = [
            "from __future__ import annotations\n\n",
            "import pathlib\n",
            "import secrets\n",
            "import typing\n\n",
            "from fntypes.co import Variative, Nothing\n",
            "from telegrinder.model import From, Model, field\n",
            "from telegrinder.types.input_file import InputFile\n",
            "from functools import cached_property\n",
            "from telegrinder.msgspec_utils.custom_types import Option, Literal, datetime, timedelta\n\n",
        ]

        if self.config.generator.objects.fields_literal_types:
            lines.append("from telegrinder.types.enums import *  # noqa: F403\n")

        all_ = ["Model"]
        for object_schema in sorted(self.objects, key=lambda obj: obj.subtypes, reverse=True):
            if object_schema.name != "InputFile":
                lines.append(self.make_object(object_schema) + "\n\n")

            all_.append(object_schema.name)

        lines.append(f"\n__all__ = {tuple(set(all_))!r}\n")

        objects_file = path / "objects.py"
        with open(file=objects_file, mode="w+", encoding="UTF-8") as f:
            f.writelines(lines)

        enums_all = tuple(
            set(
                importlib.import_module(name=(path / "enums").as_posix().replace("/", ".")).__all__,
            ),
        )
        with open(file=path / "__init__.py", mode="w+", encoding="UTF-8") as f:
            f.writelines(
                [
                    "from telegrinder.types.enums import *\n",
                    "from telegrinder.types.objects import *\n\n",
                    f"__all__ = {enums_all + tuple(all_)}\n",
                ],
            )

        logger.info(
            "Generation of {} objects into {} has been completed.",
            len(self.objects),
            objects_file,
        )


class MethodGenerator(ABCGenerator):
    def __init__(
        self,
        methods: list[MethodSchema],
        config: Config,
        parent_types: dict[str, list[str]],
        *,
        version: str,
        release_date: str,
    ) -> None:
        self.methods = methods
        self.config = config
        self.parent_types = parent_types
        self.release_date = release_date
        self.version = version

    @staticmethod
    def make_type_hint(
        types: list[str],
        parent_types: dict[str, list[str]] | None = None,
        is_return_type: bool = False,
    ) -> str:
        if not types:
            return "typing.Any"

        if len(types) == 1:
            return convert_to_python_type(types[0], parent_types)

        if len([x for x in types if x.startswith("Array of")]) > 1:
            array_of_types = [
                types.remove(v) or v.removeprefix("Array of") for v in types[:] if v.startswith("Array of")
            ]
            types.append("Array of " + ", ".join(array_of_types))

        sep = ", " if is_return_type else " | "
        return ("Variative[%s]" if is_return_type else "%s") % sep.join(
            convert_to_python_type(tp, parent_types) for tp in types
        )

    def get_param_literal_types(
        self,
        method_name: str,
        param_name: str,
        /,
    ) -> MethodsParamsLiteralTypesParam | None:
        for p_literal_types in self.config.generator.methods.params_literal_types:
            if method_name == p_literal_types.method_name:
                for param in p_literal_types.params:
                    if param.name == param_name:
                        return param

        return None

    def get_typed_default_param(self, name: str, /) -> TypedDefaultParameter | None:
        name = makesafe_name(name)

        for param in self.config.generator.api.typed_default_parameters:
            if param.name == name:
                return param

        return None

    def make_method_body(self, method_schema: MethodSchema, /) -> str:
        field_descriptions = filter(
            None,
            [
                (
                    f":param {makesafe_name(x.name)}: "
                    + chunks_str(
                        x.description + ("." if not x.description.endswith(".") else ""),
                        sep=" \\\n" + TAB + TAB,
                    ).replace('"', "`")
                    if x.description
                    else ""
                )
                for x in method_schema.params or []
            ],
        )
        return (
            '"""'
            + (
                f"Method `{method_schema.name}`, see the [documentation]({method_schema.href})\n\n{TAB * 2}"
                + chunks_str(f"{TAB}\n".join(method_schema.description or []), sep="\n" + TAB + TAB)
                + f"\n\n{TAB * 2}"
                + (f"\n\n{TAB * 2}".join(field_descriptions)).strip()
            ).strip()
            + f'\n{TAB * 2}"""\n\n'
            + '{}method_response = await self.api.request_raw("{}", get_params(locals()),)\n'.format(
                TAB * 2,
                method_schema.name,
            )
            + f"{TAB * 2}return full_result(method_response, "
            + f"{self.make_type_hint(method_schema.returns or [], self.parent_types, is_return_type=True)})"
        )

    def make_method_params(
        self,
        method_name: str,
        params: list[MethodParameter] | None = None,
    ) -> list[str]:
        if not params:
            return []

        result = []
        for p in (
            *filter(lambda x: x.required and not self.get_typed_default_param(x.name), params),
            *filter(lambda x: not x.required or self.get_typed_default_param(x.name), params),
        ):
            literal_types = self.get_param_literal_types(method_name, p.name)
            param_name = makesafe_name(p.name)

            if literal_types is not None:
                tp = literal_types.enum or "typing.Literal[%s]" % ", ".join(
                    f'"{x}"' if isinstance(x, str) else str(x) for x in literal_types.literals
                )
                tp = f"list[{tp}]" if any("Array of" in x for x in p.types) else tp
                p.types = [tp]

            elif is_timestamp_type(param_name, types=p.types):
                p.types.insert(p.types.index("Integer" if "Integer" in p.types else "Float"), "Timestamp")

            elif p.description and is_unixtime_type(param_name, p.types, p.description):
                p.types.insert(p.types.index("Integer"), "Unixtime")

            default_param_value = (
                None if self.get_typed_default_param(param_name) is None else f'default_params["{param_name}"]'
            )
            code = f"{TAB * 2}{param_name}: "

            if not p.required or default_param_value:
                code += (
                    f"{self.make_type_hint(p.types)} {'| None ' if not p.required else ''}= {default_param_value}"
                )
            else:
                code += self.make_type_hint(p.types)

            result.append(code)

        return result

    def make_return_type(self, returns: list[str] | None = None) -> str:
        if not returns:
            return "Result[bool, APIError]"
        return f"Result[{self.make_type_hint(returns, self.parent_types, is_return_type=True)}, APIError]"

    def make_method(self, method_schema: MethodSchema) -> str:
        code = (
            f"{TAB}async def {camel_to_snake(method_schema.name)}(self,"
            + ("*,\n" if method_schema.params else "")
            + ",\n".join(self.make_method_params(method_schema.name, method_schema.params))
            + ("," if method_schema.params else "")
            + f"**other: typing.Any{',' if method_schema.params else ''}) -> "
            + self.make_return_type(method_schema.returns)
            + f":\n{TAB * 2}"
            + self.make_method_body(method_schema)
        )
        return code

    def generate(self, path: pathlib.Path) -> None:
        if not self.methods:
            logger.error("Methods is empty.")
            sys.exit(-1)

        logger.debug("Generate methods...")
        docstring = '    """Telegram Bot API version `{}`, released `{}`."""\n\n'.format(
            self.version,
            self.release_date,
        )
        default_params_typeddict = 'typing.TypedDict("DefaultParams", {},)'.format(
            "{%s}"
            % ", ".join(
                f""""{x.name}": {convert_to_python_type(x.type, parent_types=self.parent_types)}"""
                for x in self.config.generator.api.typed_default_parameters
            )
        )
        lines = [
            "import typing\n",
            "from datetime import datetime, timedelta\n\n"
            "from fntypes.co import Result, Variative\n"
            "from telegrinder.api.error import APIError\n",
            "from telegrinder.types.methods_utils import ProxiedDict, full_result, get_params\n",
            "from telegrinder.types.enums import *  # noqa: F403\n"
            "from telegrinder.types.objects import *  # noqa: F403\n\n"
            "if typing.TYPE_CHECKING:\n",
            "    from telegrinder.api.api import API\n\n\n",
            "class APIMethods:\n" + docstring,
            f"\n\n    default_params = ProxiedDict({default_params_typeddict},)\n\n",
            '    def __init__(self, api: "API") -> None:\n',
            "        self.api = api\n\n",
        ]

        for method_schema in self.methods:
            lines.append(self.make_method(method_schema) + "\n\n")

        methods_file = path / "methods.py"
        with open(file=methods_file, mode="w+", encoding="UTF-8") as f:
            lines.append('\n\n\n__all__ = ("APIMethods",)\n\n')
            f.writelines(lines)

        logger.info(
            "Generation of {} methods into {} has been completed.",
            len(self.methods),
            methods_file,
        )


def generate(
    *,
    directory: str | pathlib.Path | None = None,
    object_generator: ABCGenerator | None = None,
    method_generator: ABCGenerator | None = None,
) -> None:
    logger.debug("Reading config file...")
    config_toml = read_config(TYPEGEN_DIR / "config.toml")
    config = config_toml.as_model(Config, dec_hook=dec_hook)

    directory = pathlib.Path(directory) if directory else config.generator.directory
    if not os.path.exists(directory):
        logger.warning(f"Directory {directory} not found. Make directory...")
        os.makedirs(directory)

    logger.info(f"Generation... Directory: {directory}")

    schema = download_schema(config_toml, config)
    if object_generator is None or method_generator is None:
        object_generator = object_generator or ObjectGenerator(
            objects=schema.objects,
            config=config,
        )
        method_generator = method_generator or MethodGenerator(
            methods=schema.methods,
            parent_types=object_generator.parent_types if isinstance(object_generator, ObjectGenerator) else {},
            config=config,
            version=schema.version.split()[-1],
            release_date=schema.release_date,
        )

    object_generator.generate(directory)
    method_generator.generate(directory)
    logger.info("Successful generation!")

    run_ruff_formatter(directory)
    merge_shortcuts(path_api_methods=directory / "methods.py")


__all__ = ("ABCGenerator", "MethodGenerator", "ObjectGenerator", "generate")



import dataclasses
import typing

import msgspec

from telegrinder.node.base import IsNode, Node, as_node
from telegrinder.tools.magic.annotations import Annotations
from telegrinder.tools.magic.dictionary import extract


class Collection(Node):
    __subnodes__: typing.ClassVar[dict[str, IsNode] | None] = None

    @classmethod
    def get_subnodes(cls) -> dict[str, IsNode]:
        if cls.__subnodes__ is None:
            cls.__subnodes__ = {
                name: node
                for name, annotation in Annotations(obj=cls).get(cache=True, exclude_forward_refs=True).items()
                if (node := as_node(annotation, raise_exception=False)) is not None
            }

        return cls.__subnodes__

    @classmethod
    def compose(cls, **kwargs: typing.Any) -> typing.Self:
        nodes = extract(cls.__subnodes__ or (), kwargs)

        if dataclasses.is_dataclass(cls) or issubclass(cls, msgspec.Struct):
            return cls(**nodes)

        instance = cls()
        for name, value in nodes.items():
            setattr(instance, name, value)

        return instance


__all__ = ("Collection",)



import typing

from telegrinder.api.api import API
from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.bot.dispatch.waiter_machine.hasher.hasher import Hasher
from telegrinder.bot.scenario.checkbox import Checkbox, ChoiceAction


class Choice[Key: typing.Hashable](Checkbox[Key]):
    async def handle(self, cb: CallbackQueryCute) -> bool:
        code = cb.data.unwrap().replace(self.random_code + "/", "", 1)
        if code == ChoiceAction.READY:
            return False

        for choice in self.choices:
            choice.is_picked = False

        for i, choice in enumerate(self.choices):
            if choice.code == code:
                self.choices[i].is_picked = True
                await cb.ctx_api.edit_message_text(
                    text=self.message,
                    chat_id=cb.message.unwrap().v.chat.id,
                    message_id=cb.message.unwrap().v.message_id,
                    parse_mode=self.PARSE_MODE,
                    reply_markup=self.get_markup(),
                )

        return True

    async def wait(self, hasher: Hasher[CallbackQueryCute, int], api: API) -> tuple[Key, int]:
        if len(tuple(choice for choice in self.choices if choice.is_picked)) != 1:
            raise ValueError("Exactly one choice must be picked.")

        choices, m_id = await super().wait(hasher, api)
        return tuple(choices.keys())[tuple(choices.values()).index(True)], m_id


__all__ = ("Choice",)



"""Modern visionary telegram bot framework.

* Type hinted
* Customizable and extensible
* Ready to use scenarios and rules
* Fast models built on [msgspec](https://github.com/jcrist/msgspec)
* Both low-level and high-level API

Basic example:

```python
from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import Text

logger.set_level("INFO")
api = API(token=Token("123:token"))
bot = Telegrinder(api)


@bot.on.message(Text("/start"))
async def start(message: Message) -> None:
    me = (await api.get_me()).unwrap()
    await message.answer(f"Hello, {message.from_user.full_name}! I'm {me.full_name}.")


bot.run_forever()
```
"""

import typing

from .api import API, APIError, APIResponse, APIServerError, Token
from .bot import (
    CALLBACK_QUERY_FOR_MESSAGE,
    CALLBACK_QUERY_FROM_CHAT,
    CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE,
    MESSAGE_FROM_USER,
    MESSAGE_FROM_USER_IN_CHAT,
    MESSAGE_IN_CHAT,
    ABCDispatch,
    ABCHandler,
    ABCMiddleware,
    ABCPolling,
    ABCReturnManager,
    ABCRule,
    ABCScenario,
    ABCView,
    AudioReplyHandler,
    BaseCute,
    BaseReturnManager,
    BaseView,
    CallbackQueryCute,
    CallbackQueryReturnManager,
    CallbackQueryView,
    ChatJoinRequestCute,
    ChatJoinRequestView,
    ChatMemberUpdatedCute,
    ChatMemberView,
    Checkbox,
    Choice,
    Context,
    Dispatch,
    DocumentReplyHandler,
    FuncHandler,
    Hasher,
    InlineQueryCute,
    InlineQueryReturnManager,
    MediaGroupReplyHandler,
    MessageCute,
    MessageReplyHandler,
    MessageReturnManager,
    MessageView,
    PhotoReplyHandler,
    Polling,
    PreCheckoutQueryCute,
    PreCheckoutQueryManager,
    PreCheckoutQueryView,
    RawEventView,
    ShortState,
    StickerReplyHandler,
    Telegrinder,
    UpdateCute,
    VideoReplyHandler,
    ViewBox,
    WaiterMachine,
    action,
    register_manager,
)
from .client import ABCClient, AiohttpClient
from .model import Model, field
from .modules import logger
from .tools.formatting import HTMLFormatter
from .tools.global_context import ABCGlobalContext, GlobalContext, TelegrinderContext
from .tools.input_file_directory import InputFileDirectory
from .tools.keyboard import (
    ABCKeyboard,
    ABCStaticKeyboard,
    BaseButton,
    BaseKeyboard,
    BaseStaticButton,
    BaseStaticKeyboard,
    Button,
    InlineButton,
    InlineKeyboard,
    Keyboard,
    RowButtons,
    StaticButton,
    StaticInlineButton,
    StaticInlineKeyboard,
    StaticKeyboard,
)
from .tools.lifespan import Lifespan
from .tools.loop_wrapper import DelayedTask, LoopWrapper
from .tools.parse_mode import ParseMode
from .tools.state_storage import ABCStateStorage, MemoryStateStorage, StateData

Update: typing.TypeAlias = UpdateCute
Message: typing.TypeAlias = MessageCute
PreCheckoutQuery: typing.TypeAlias = PreCheckoutQueryCute
ChatJoinRequest: typing.TypeAlias = ChatJoinRequestCute
ChatMemberUpdated: typing.TypeAlias = ChatMemberUpdatedCute
CallbackQuery: typing.TypeAlias = CallbackQueryCute
InlineQuery: typing.TypeAlias = InlineQueryCute
Bot: typing.TypeAlias = Telegrinder


__all__ = (
    "API",
    "CALLBACK_QUERY_FOR_MESSAGE",
    "CALLBACK_QUERY_FROM_CHAT",
    "CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE",
    "MESSAGE_FROM_USER",
    "MESSAGE_FROM_USER_IN_CHAT",
    "MESSAGE_IN_CHAT",
    "ABCClient",
    "ABCDispatch",
    "ABCGlobalContext",
    "ABCHandler",
    "ABCKeyboard",
    "ABCMiddleware",
    "ABCPolling",
    "ABCReturnManager",
    "ABCRule",
    "ABCScenario",
    "ABCStateStorage",
    "ABCStateStorage",
    "ABCStaticKeyboard",
    "ABCView",
    "APIError",
    "APIResponse",
    "APIServerError",
    "AiohttpClient",
    "AudioReplyHandler",
    "BaseButton",
    "BaseCute",
    "BaseKeyboard",
    "BaseReturnManager",
    "BaseStaticButton",
    "BaseStaticKeyboard",
    "BaseView",
    "Bot",
    "Button",
    "CallbackQuery",
    "CallbackQueryCute",
    "CallbackQueryReturnManager",
    "CallbackQueryView",
    "ChatJoinRequest",
    "ChatJoinRequestCute",
    "ChatJoinRequestView",
    "ChatMemberUpdated",
    "ChatMemberUpdatedCute",
    "ChatMemberView",
    "Checkbox",
    "Choice",
    "Context",
    "DelayedTask",
    "Dispatch",
    "DocumentReplyHandler",
    "FuncHandler",
    "GlobalContext",
    "HTMLFormatter",
    "Hasher",
    "InlineButton",
    "InlineKeyboard",
    "InlineQuery",
    "InlineQueryCute",
    "InlineQueryReturnManager",
    "InputFileDirectory",
    "Keyboard",
    "Lifespan",
    "LoopWrapper",
    "MediaGroupReplyHandler",
    "MemoryStateStorage",
    "MemoryStateStorage",
    "Message",
    "MessageCute",
    "MessageReplyHandler",
    "MessageReplyHandler",
    "MessageReturnManager",
    "MessageView",
    "Model",
    "ParseMode",
    "PhotoReplyHandler",
    "Polling",
    "PreCheckoutQuery",
    "PreCheckoutQueryCute",
    "PreCheckoutQueryManager",
    "PreCheckoutQueryView",
    "RawEventView",
    "RowButtons",
    "ShortState",
    "StateData",
    "StateData",
    "StaticButton",
    "StaticInlineButton",
    "StaticInlineKeyboard",
    "StaticKeyboard",
    "StickerReplyHandler",
    "Telegrinder",
    "TelegrinderContext",
    "Token",
    "Update",
    "UpdateCute",
    "VideoReplyHandler",
    "ViewBox",
    "WaiterMachine",
    "action",
    "field",
    "logger",
    "register_manager",
)



import typing

from telegrinder.node.base import IsNode, Node


class ContainerNode(Node):
    linked_nodes: typing.ClassVar[list[IsNode]]
    composer: typing.Callable[..., typing.Awaitable[typing.Any]]

    __subnodes__: typing.ClassVar[dict[str, IsNode] | None] = None

    @classmethod
    async def compose(cls, **kw: typing.Any) -> typing.Any:
        subnodes = cls.get_subnodes().keys()
        return await cls.composer(*tuple(t[1] for t in sorted(kw.items(), key=lambda t: t[0]) if t[0] in subnodes))

    @classmethod
    def get_subnodes(cls) -> dict[str, IsNode]:
        if cls.__subnodes__ is None:
            cls.__subnodes__ = {f"node_{i}": node_t for i, node_t in enumerate(cls.linked_nodes, start=1)}
        return cls.__subnodes__

    @classmethod
    def link_nodes(
        cls,
        linked_nodes: list[IsNode],
        composer: typing.Callable[..., typing.Awaitable[typing.Any]],
    ) -> type[typing.Self]:
        return type(
            cls.__name__,
            (cls,),
            {"linked_nodes": linked_nodes, "composer": classmethod(composer), "__module__": __name__},
        )  # type: ignore


__all__ = ("ContainerNode",)



import dataclasses
import types
import typing

from telegrinder.bot.rules.abc import ABCRule
from telegrinder.bot.rules.payload import PayloadJsonEqRule, PayloadMarkupRule, PayloadModelRule
from telegrinder.bot.rules.text import Text
from telegrinder.tools.callback_data_serialization import ABCDataSerializer
from telegrinder.tools.keyboard.buttons.base import BaseStaticButton
from telegrinder.tools.keyboard.buttons.buttons import Button, InlineButton


@dataclasses.dataclass
class StaticButtonMixin(BaseStaticButton):
    __and__: typing.ClassVar[None] = None
    __or__: typing.ClassVar[None] = None


@dataclasses.dataclass
class StaticButton(StaticButtonMixin, Button, Text):
    if not typing.TYPE_CHECKING:

        def __init__(self, text, **kwargs):
            self.row = kwargs.pop("row", False)
            Button.__init__(self, text=text, **kwargs)
            Text.__init__(self, text)


@dataclasses.dataclass
class StaticInlineButton(StaticButtonMixin, InlineButton, ABCRule):
    if not typing.TYPE_CHECKING:

        def __init__(self, *args, **kwargs):
            self.row = kwargs.pop("row", False)
            InlineButton.__init__(self, *args, **kwargs)

        check = lambda self, *args, **kwargs: False
    else:

        def check(self, *args: typing.Any, **kwargs: typing.Any) -> bool: ...

    def __post_init__(self, callback_data_serializer: ABCDataSerializer[typing.Any] | None) -> None:
        if isinstance(self.callback_data, str):
            self.check = PayloadMarkupRule(self.callback_data).check
        elif isinstance(self.callback_data, dict):
            self.check = PayloadJsonEqRule(self.callback_data).check
        elif not isinstance(self.callback_data, (bytes, types.NoneType)):
            self.check = PayloadModelRule(
                type(self.callback_data),
                serializer=type(callback_data_serializer) if callback_data_serializer else None,
            ).check

        super().__post_init__(callback_data_serializer)


__all__ = ("StaticButton", "StaticInlineButton")



import typing


def separate_by_plus_char(iterable: typing.Iterable[str] | None = None, /) -> str | None:
    return None if not iterable else "+".join(iterable)


__all__ = ("separate_by_plus_char",)



import typing

from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.types.enums import MessageEntityType
from telegrinder.types.objects import MessageEntity

type Entity = str | MessageEntityType

Message: typing.TypeAlias = MessageCute


class HasEntities(ABCRule):
    def check(self, message: Message) -> bool:
        return bool(message.entities)


class MessageEntities(ABCRule, requires=[HasEntities()]):
    def __init__(self, entities: Entity | list[Entity], /) -> None:
        self.entities = [entities] if not isinstance(entities, list) else entities

    def check(self, message: Message, ctx: Context) -> bool:
        message_entities: list[MessageEntity] = []
        for entity in message.entities.unwrap():
            for entity_type in self.entities:
                if entity_type == entity.type:
                    message_entities.append(entity)

        if not message_entities:
            return False

        ctx.message_entities = message_entities
        return True


__all__ = ("HasEntities", "MessageEntities")



import pathlib
import typing

import msgspec


def dec_hook(x: type[typing.Any], obj: typing.Any) -> typing.Any:
    if issubclass(x, pathlib.Path) and isinstance(obj, str | pathlib.Path):
        return x(obj)
    raise TypeError


class ObjectsIdByDefaultField(msgspec.Struct):
    name: str
    nbytes: int


class ObjectsIdByDefault(msgspec.Struct):
    object_name: str
    fields: list[ObjectsIdByDefaultField]


class ObjectsFieldsLiteralTypesField(msgspec.Struct):
    name: str
    literals: list[str | int] = msgspec.field(default_factory=lambda: [])
    enum: str | None = None
    default: str | None = None

    @property
    def enum_default(self) -> str | None:
        if self.enum is not None and self.default is not None:
            return f"{self.enum}.{self.default}"
        return None

    @property
    def literals_default(self) -> str | None:
        if self.enum is None and self.default is not None:
            return self.default
        return None


class ObjectsFieldsLiteralTypes(msgspec.Struct):
    object_name: str
    fields: list[ObjectsFieldsLiteralTypesField]


class GeneratorObjects(msgspec.Struct):
    id_by_default: list[ObjectsIdByDefault] = msgspec.field(name="id-by-default")
    fields_literal_types: list[ObjectsFieldsLiteralTypes] = msgspec.field(name="fields-literal-types")


class MethodsParamsLiteralTypesParam(ObjectsFieldsLiteralTypesField):
    pass


class MethodsParamsLiteralTypes(msgspec.Struct):
    method_name: str
    params: list[MethodsParamsLiteralTypesParam]


class GeneratorMethods(msgspec.Struct):
    params_literal_types: list[MethodsParamsLiteralTypes] = msgspec.field(name="params-literal-types")


class TypedDefaultParameter(msgspec.Struct):
    name: str
    type: str


class GeneratorAPI(msgspec.Struct):
    typed_default_parameters: list[TypedDefaultParameter] = msgspec.field(name="typed-default-parameters")


class Generator(msgspec.Struct):
    api: GeneratorAPI
    objects: GeneratorObjects
    methods: GeneratorMethods
    directory: pathlib.Path
    nicifications_path: pathlib.Path | None = None


class TelegramBotAPI(msgspec.Struct):
    version: str
    schema_url: str

    @property
    def version_number(self) -> float:
        return float(self.version.removeprefix("v"))


class Config(msgspec.Struct):
    telegram_bot_api: TelegramBotAPI = msgspec.field(name="telegram-bot-api")
    generator: Generator


__all__ = ("Config",)



from __future__ import annotations

import typing
from functools import cached_property

from fntypes.misc import from_optional
from fntypes.option import Option

from telegrinder.bot.cute_types import BaseCute
from telegrinder.bot.dispatch.view.base import BaseView

type HasherWithData[Event: BaseCute, Data] = tuple[Hasher[Event, Data], Data]

Event = typing.TypeVar("Event", bound=BaseCute, covariant=True)
Data = typing.TypeVar("Data", covariant=True)


def ECHO[T](__x: T) -> T:  # noqa
    return __x


class Hasher(typing.Generic[Event, Data]):
    def __init__(
        self,
        view_class: type[BaseView],
        get_hash_from_data: typing.Callable[[Data], typing.Hashable | None] | None = None,
        get_data_from_event: typing.Callable[[Event], Data | None] | None = None,
    ) -> None:
        self.view_class = view_class
        self._get_hash_from_data = get_hash_from_data
        self._get_data_from_event = get_data_from_event

    def __call__[D](self: "Hasher[Event, D]", data: D, /) -> HasherWithData[Event, D]:
        return (self, data)

    def __hash__(self) -> int:
        return hash(self.name)

    def __repr__(self) -> str:
        return f"<Hasher {self.name}>"

    @cached_property
    def name(self) -> str:
        return f"{self.view_class.__name__}_{id(self)}"

    def get_hash_from_data[D](self: "Hasher[Event, D]", data: D) -> Option[typing.Hashable]:
        if self._get_hash_from_data is None:
            raise NotImplementedError
        return from_optional(self._get_hash_from_data(data))

    def get_data_from_event[E: BaseCute](self: "Hasher[E, Data]", event: E) -> Option[Data]:
        if not self._get_data_from_event:
            raise NotImplementedError
        return from_optional(self._get_data_from_event(event))

    def get_hash_from_data_from_event[E: BaseCute](
        self: "Hasher[E, Data]",
        event: E,
    ) -> Option[typing.Hashable]:
        return self.get_data_from_event(event).then(self.get_hash_from_data)  # type: ignore


__all__ = ("Hasher",)



from telegrinder.msgspec_utils.custom_types.datetime import datetime, timedelta
from telegrinder.msgspec_utils.custom_types.enum_meta import BaseEnumMeta
from telegrinder.msgspec_utils.custom_types.literal import Literal
from telegrinder.msgspec_utils.custom_types.option import Option

__all__ = ("BaseEnumMeta", "Literal", "Option", "datetime", "timedelta")



from __future__ import annotations

import re
import typing

from fntypes.option import Nothing, Option
from vbml.patcher.abc import ABCPatcher
from vbml.patcher.patcher import Patcher

from telegrinder.tools.global_context.global_context import GlobalContext, ctx_var, runtime_init
from telegrinder.tools.loop_wrapper import LoopWrapper

if typing.TYPE_CHECKING:
    from telegrinder.node.composer import Composer


@runtime_init
class TelegrinderContext(GlobalContext):
    """The type-hinted telegrinder context called `telegrinder`.

    Example:
    ```
    from telegrinder.tools.global_context import GlobalContext, TelegrinderContext

    telegrinder_ctx = TelegrinderContext(...)  # with type-hints
    assert telegrinder_ctx == GlobalContext("telegrinder")  # ok
    ```

    """

    __ctx_name__ = "telegrinder"

    composer: Option[Composer] = ctx_var(default=Nothing(), init=False)
    vbml_pattern_flags: re.RegexFlag | None = ctx_var(default=None, init=False)
    vbml_patcher: ABCPatcher = ctx_var(default_factory=Patcher, init=False)
    loop_wrapper: LoopWrapper = ctx_var(default_factory=LoopWrapper, init=False)


__all__ = ("TelegrinderContext",)
